from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import sys
import os

import subprocess
import shutil

# Check for ffmpeg on import - REQUIRED system dependency
try:
    if shutil.which("ffmpeg") is None:
        raise FileNotFoundError
    subprocess.run(["ffmpeg", "-version"], capture_output=True, check=True)
except (FileNotFoundError, subprocess.CalledProcessError):
    raise RuntimeError(
        "ffmpeg is a required system requirement but not found. Install with:\n"
        "conda install ffmpeg=6.1.2 -c conda-forge\n"
        "or visit: https://ffmpeg.org/download.html"
    )

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'models/clipscore_models/InternVideo2/multi_modality/'))

from .constants import HF_CACHE_DIR
from .vqascore import VQAScore, list_all_vqascore_models
from .clipscore import CLIPScore, list_all_clipscore_models
from .itmscore import ITMScore, list_all_itmscore_models


def list_all_models():
    return list_all_vqascore_models() + list_all_clipscore_models() + list_all_itmscore_models()

def get_score_model(model='clip-flant5-xxl', device='cuda', cache_dir=HF_CACHE_DIR, **kwargs):
    if model in list_all_vqascore_models():
        return VQAScore(model, device=device, cache_dir=cache_dir, **kwargs)
    elif model in list_all_clipscore_models():
        return CLIPScore(model, device=device, cache_dir=cache_dir, **kwargs)
    elif model in list_all_itmscore_models():
        return ITMScore(model, device=device, cache_dir=cache_dir, **kwargs)
    else:
        raise NotImplementedError()


# Copyright (c) Meta Platforms, Inc. and affiliates.

# This file from the xFormers repo is just a example of how to implement
# probing of the activations of a model, without changing anything.
# By default, the linear inputs/outputs/gradients are logged, as well as
# the attention logits+entropy. It is possible to log an additional tensor, eg:
# x = log_stats(x, "name")
#
# Known limitations:
# * Only a subset of the attention biases is supported
# * Torch-compile is disabled automatically when this is enabled
# * Only tested with bf16/f16/f32 datatypes

import contextlib
import functools
import json
import math
import os
import uuid
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import (
    CheckpointImpl, checkpoint_wrapper)
from torch.fx.operator_schemas import normalize_function
from torch.nn.attention import SDPBackend, sdpa_kernel
from torch.utils._python_dispatch import TorchDispatchMode
from torch.utils._pytree import tree_map
from torch.utils.module_tracker import ModuleTracker
from xformers.ops import fmha


@torch.library.custom_op("torchprobe::log", mutates_args=(), device_types=None)
def _log(x: torch.Tensor, name: str, uid: str) -> None:
    pass


@_log.register_fake
def _log_fake(x: torch.Tensor, name: str, uid: str) -> None:
    pass


class _LogStats(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x: torch.Tensor, name: str):
        uid = str(uuid.uuid4())
        torch.ops.torchprobe.log(x, name, uid)
        ctx.name = name
        ctx.uid = uid
        return x

    @staticmethod
    def backward(ctx, grad: torch.Tensor):
        torch.ops.torchprobe.log(grad, f"{ctx.name}.g", ctx.uid)
        return grad, None


_PROBING_ENABLED = False


def log_stats(x: torch.Tensor, name: str) -> torch.Tensor:
    if not _PROBING_ENABLED:
        return x
    return _LogStats.apply(x, name)


QUANTILES = [
    0.0000001,
    0.000001,
    0.00001,
    0.0001,
    0.001,
    0.01,
    0.05,
    0.1,
    0.3,
    0.5,
    0.7,
    0.9,
    0.95,
    0.99,
    0.999,
    0.9999,
    0.99999,
    0.999999,
    0.9999999,
]


@functools.cache
def _get_quantiles(device: torch.device, dtype) -> torch.Tensor:
    return torch.tensor(QUANTILES, device=device, dtype=dtype)


def _get_stats(x_: torch.Tensor, remove_inf=False) -> Dict[str, Any]:
    if x_.dtype not in [torch.float, torch.double, torch.float16, torch.bfloat16]:
        return {}
    x = x_.flatten()
    if remove_inf:
        x = x[x.abs() < float("inf")]
    if x.dtype is not torch.double:
        x = x.float()
    xabs = x.abs()
    quantiles = _get_quantiles(x.device, x.dtype)
    mean = x.mean()
    std = x.std()
    return {
        "shape": tuple(x_.shape),
        "mean": mean,
        "std": std,
        "skew": (((x - mean) / std) ** 3).double().mean(),
        "kurtosis": (((x - mean) / std) ** 4).double().mean(),
        "abs.mean": xabs.mean(),
        "max": x.max(),
        "min": x.min(),
        # Note: `quantile` takes at most 2**24 elements, see
        # https://github.com/pytorch/pytorch/issues/64947
        "quantiles": torch.quantile(x[: 2**24], quantiles),
    }


def _mask_attn_causal_inplace(logits: torch.Tensor, q_idx, q_len, kv_len) -> None:
    assert logits.ndim == 4
    logits[:, :, :, q_idx + kv_len - q_len + 1 :] = -math.inf


def _mask_attn_logits(
    logits: torch.Tensor,
    q_idx: List[int],
    *,
    causal: bool,
    cu_seqlens_q: Optional[torch.Tensor] = None,
    cu_seqlens_k: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    assert logits.dtype is torch.float32
    # Handle BlockDiagonalMask
    if cu_seqlens_q is not None:
        assert cu_seqlens_k is not None
        # Expect BHMqMkv
        assert logits.ndim == 4, logits.shape
        qs = cu_seqlens_q.tolist()
        ks = cu_seqlens_k.tolist()
        q_batchid = []
        k_batchid = [-2] * logits.shape[-1]
        q_idx_i = 0
        for bid, (q0, q1, k0, k1) in enumerate(zip(qs, qs[1:], ks, ks[1:])):
            for k in range(k0, k1):
                k_batchid[k] = bid
            while q_idx_i < len(q_idx) and q_idx[q_idx_i] < q1:
                q_batchid.append(bid)
                if causal:
                    _mask_attn_causal_inplace(
                        logits[:, :, q_idx_i : q_idx_i + 1, k0:k1],
                        q_idx[q_idx_i] - q0,
                        q1 - q0,
                        k1 - k0,
                    )
                q_idx_i += 1
        mask_out = (
            torch.tensor(q_batchid, device=logits.device)[None, None, :, None]
            != torch.tensor(k_batchid, device=logits.device)[None, None, None, :]
        )
        logits[mask_out.expand_as(logits)] = -math.inf
        assert q_idx_i == len(q_idx)
    elif causal:
        for q_idx_i in range(len(q_idx)):
            _mask_attn_causal_inplace(
                logits[:, :, q_idx_i : q_idx_i + 1, :],
                q_idx[q_idx_i],
                logits.shape[2],
                logits.shape[3],
            )
    return logits


def _attn_queries_subset(num_queries: int) -> List[int]:
    return list(range(0, num_queries, max(1, num_queries // 128)))


@torch.no_grad()
def _compute_attn_stats_sdpa(
    probe,
    path: str,
    # supports arguments both cudnn + flash backends
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask=None,
    attn_bias=None,
    dropout_p=0.0,
    is_causal=False,
    scale=None,
    compute_log_sumexp=True,
    return_debug_mask=False,
    **kwargs,
):
    if scale is None:
        scale = 1 / (query.shape[-1] ** 0.5)
    # Filter-out not supported cases
    if attn_mask is not None or attn_bias is not None or dropout_p != 0.0 or kwargs:
        probe.store[f"{path}::attn"] = {
            "query.shape": tuple(query.shape),
            "key.shape": tuple(key.shape),
            "value.shape": tuple(value.shape),
            "attn_mask": attn_mask.shape if attn_mask is not None else None,
            "dropout_p": dropout_p,
            "is_causal": is_causal,
            "scale": scale,
            "unk_kwargs": list(kwargs.keys()),
        }
        return
    # Take a subset of the queries and compute the logits
    query_s = _attn_queries_subset(query.shape[-2])
    logits = query[:, :, query_s] @ key.transpose(-1, -2) * scale
    logits = _mask_attn_logits(logits.float(), query_s, causal=is_causal)
    p = logits.float().softmax(-1)
    masked_logsoft = logits.log_softmax(-1).where(
        (logits > -math.inf), torch.zeros_like(logits)
    )
    entropy = -(p * masked_logsoft).sum(-1)
    probe.log_tensor(f"{path}::attn_entropy", entropy)
    probe.log_tensor(f"{path}::attn_logits", logits, remove_inf=True)


@torch.no_grad()
def _compute_attn_stats_flash(
    probe,
    path: str,
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    cu_seqlens_q: Optional[torch.Tensor],
    cu_seqlens_k: Optional[torch.Tensor],
    seqused_k: Optional[torch.Tensor],
    max_seqlen_q: int,
    max_seqlen_k: int,
    p: float,
    softmax_scale: float,
    is_causal: bool,
    window_left: int,
    window_right: int,
    return_softmax: bool,
    block_tables: Optional[torch.Tensor],
    unpadded_lse: bool = False,
) -> None:
    # Filter-out not supported cases
    if (
        seqused_k is not None
        or p != 0.0
        or window_left >= 0
        or window_right >= 0
        or block_tables is not None
    ):
        probe.store[f"{path}::attn"] = {
            "query.shape": tuple(query.shape),
            "key.shape": tuple(key.shape),
            "value.shape": tuple(value.shape),
            "op": "flash",
        }
        return

    if cu_seqlens_q is not None:
        assert query.ndim == 3, query.shape
        query, key, value = query[None], key[None], value[None]
    assert query.ndim == 4, query.shape

    # Take a subset of the queries and compute the logits
    query_s = _attn_queries_subset(query.shape[1])
    logits = (
        query[:, query_s].transpose(1, 2)
        @ key.transpose(1, 2).transpose(-1, -2)
        * softmax_scale
    )
    logits = _mask_attn_logits(
        logits.float(),
        query_s,
        cu_seqlens_q=cu_seqlens_q,
        cu_seqlens_k=cu_seqlens_k,
        causal=is_causal,
    )
    p = logits.float().softmax(-1)
    masked_logsoft = logits.log_softmax(-1).where(
        (logits > -math.inf), torch.zeros_like(logits)
    )
    entropy = -(p * masked_logsoft).sum(-1)
    probe.log_tensor(f"{path}::attn_entropy", entropy)
    probe.log_tensor(f"{path}::attn_logits", logits, remove_inf=True)


def _tensors_to_python(x):
    if not isinstance(x, torch.Tensor):
        return x
    return x.tolist()


# class syntax
class LinearBwType(Enum):
    DW = 1
    DX = 2
    UNKNOWN = 3


class AutoProbeD(TorchDispatchMode):
    def __init__(self, module: nn.Module, write_file: Optional[str] = None) -> None:
        self.write_file = Path(write_file) if write_file is not None else None
        self.write_tensors_tmpdir: Optional[Path] = None
        self.compile_disabler = TorchCompileDisabler(module)
        self.mod_tracker = ModuleTracker()
        self.count_per_path: Dict[str, int] = defaultdict(int)
        self.store: Dict[str, Dict[str, Any]] = {}
        self.linear_data: Dict[str, Tuple[Any, Any, Any, Any, Any]] = {}
        self.uid_to_path: Dict[str, str] = {}
        self.metadata: Any = None
        self.enabled = False
        self.verbose = bool(int(os.environ.get("PROBE_VERBOSE", "0")))

    def __enter__(self):
        global _PROBING_ENABLED
        assert not self.enabled, "Entered probe twice"
        self.compile_disabler.__enter__()
        self.mod_tracker.__enter__()
        super().__enter__()
        self.enabled = True
        _PROBING_ENABLED = True
        # self._setup_tensors_logging()
        return self

    def __exit__(self, *args) -> None:
        global _PROBING_ENABLED
        assert self.enabled, "Exiting probe without entering it"
        super().__exit__(*args)
        self.mod_tracker.__exit__(*args)
        self.compile_disabler.__exit__(*args)
        self._flush_and_clear()
        _PROBING_ENABLED = False
        self.enabled = False

    def _setup_tensors_logging(self):
        if self.write_file is not None:
            self.write_file.parent.mkdir(exist_ok=True)
            self.write_tensors_tmpdir = (
                self.write_file.parent
                / f"{self.write_file.name}-tmp-{str(uuid.uuid4())[:8]}"
            )
            self.write_tensors_tmpdir.mkdir(exist_ok=True)

    def _flush_and_clear(self) -> None:
        if self.write_file is not None:
            dump_data = tree_map(_tensors_to_python, self.store)
            with self.write_file.open("a") as fd:
                json.dump(
                    {
                        "data": dump_data,
                        "meta": self.metadata,
                        "version": 2,
                        "quantiles": QUANTILES,
                    },
                    fd,
                )
                fd.write("\n")
        if self.write_tensors_tmpdir is not None:
            assert self.write_file is not None
            dump_dir = self.write_tensors_tmpdir.parent / f"{self.write_file.name}-dump"
            dump_dir.mkdir(exist_ok=True)
            dir_name = ""
            if "it" in self.metadata:
                dir_name = f"it{int(self.metadata['it']):010}"
            if dir_name == "" or (dump_dir / dir_name).exists():
                num_files = len(list(dump_dir.glob(f"{dir_name}v*")))
                dir_name = f"{dir_name}v{num_files}"
            dump_dir = dump_dir / dir_name
            assert not dump_dir.exists()
            self.write_tensors_tmpdir.rename(dump_dir)
            self.write_tensors_tmpdir = None
        self.store.clear()
        self.count_per_path.clear()
        self.uid_to_path.clear()

    def _find_bw_path_and_type(
        self, path: str, out: torch.Tensor, args
    ) -> Tuple[str, LinearBwType]:
        """
        We are in the BW pass, and process a GEMM.
        Let's figure out:
        (1) The path for the FW pass (might differ in case of ModuleTracker bug)
        (2) The type of BW pass (eg `dw` or `dx`)
        """

        def _is_path_correct_dw(path: str) -> bool:
            # dW.t = dY.t @ X
            in_shape, w_shape, out_shape, input_sm, weight_sm = self.linear_data[path]
            return out.shape == (w_shape[1], w_shape[0]) and torch.allclose(
                input_sm, args[1][:4, :4]
            )

        def _is_path_correct_dx(path: str) -> bool:
            # dX = dY @ W.t
            in_shape, w_shape, out_shape, input_sm, weight_sm = self.linear_data[path]
            return out.shape == in_shape and torch.allclose(weight_sm, args[1][:4, :4])

        if path in self.linear_data:
            if _is_path_correct_dw(path):
                return path, LinearBwType.DW
            if _is_path_correct_dx(path):
                return path, LinearBwType.DX
        for candidate_path in self.mod_tracker.parents:
            if candidate_path not in self.linear_data:
                continue
            if _is_path_correct_dw(candidate_path):
                return candidate_path, LinearBwType.DW
            if _is_path_correct_dx(candidate_path):
                return candidate_path, LinearBwType.DX
        return path, LinearBwType.UNKNOWN

    def log_tensor(self, name: str, x: torch.Tensor, **kwargs) -> None:
        self.store[name] = _get_stats(x, **kwargs)
        if self.write_tensors_tmpdir is not None:
            name_safe = name.replace("::", "__").replace("/", "")
            torch.save(x, self.write_tensors_tmpdir / f"{name_safe}.pkl")

    def __torch_dispatch__(self, func, types, args=(), kwargs=None):
        kwargs = kwargs if kwargs else {}
        path = None
        # Find longest path
        for p in self.mod_tracker.parents:
            if p == "Global":
                continue
            if path is None or len(p) > len(path):
                path = p
        if path is None:
            path = "Global"
        path = path.replace("._checkpoint_wrapped_module", "")
        out = func(*args, **kwargs)

        # Handle linear layers
        if func._overloadpacket in [torch.ops.aten.addmm, torch.ops.aten.mm]:
            weight: torch.Tensor
            input: torch.Tensor
            if not self.mod_tracker.is_bw:
                # (technically, weight is transposed)
                if func._overloadpacket == torch.ops.aten.addmm:
                    _bias, input, weight = args[:3]
                else:
                    assert func._overloadpacket == torch.ops.aten.mm
                    input, weight = args[:2]
                self.log_tensor(f"{path}::in", input)
                self.log_tensor(f"{path}::w", weight)
                self.log_tensor(f"{path}::out", out)
                self.linear_data[path] = (
                    input.shape,
                    weight.shape,
                    out.shape,
                    input[:4, :4].clone(),
                    weight[:4, :4].T.clone(),
                )
            elif func._overloadpacket == torch.ops.aten.mm:
                # XXX: Try to find the actual path for the linear layer
                # This is messed with with Francisco's FSDP sometimes
                new_path, bwtype = self._find_bw_path_and_type(path, out, args)
                if new_path != path:
                    if self.verbose:
                        print(f"E: Fixing path `{path}` -> `{new_path}")
                    path = new_path

                if bwtype == LinearBwType.DW:
                    # dW.t = dY.t @ X
                    self.log_tensor(f"{path}::w.g", out)
                elif bwtype == LinearBwType.DX:
                    # dX = dY @ W.t
                    self.log_tensor(f"{path}::in.g", out)
                    self.log_tensor(f"{path}::out.g", args[0])
        elif func._overloadpacket in [
            torch.ops.aten._scaled_dot_product_flash_attention,
            torch.ops.aten._scaled_dot_product_cudnn_attention,
        ]:
            _, kwargs = normalize_function(
                func, args=args, kwargs=kwargs, normalize_to_only_use_kwargs=True
            )
            _compute_attn_stats_sdpa(self, path, **kwargs)
        elif func._overloadpacket == fmha.flash.FwOp.OPERATOR:
            _, kwargs = normalize_function(
                func, args=args, kwargs=kwargs, normalize_to_only_use_kwargs=True
            )
            _compute_attn_stats_flash(self, path, **kwargs)
        elif func._overloadpacket == torch.ops.torchprobe.log:
            uid = args[2]
            path = self.uid_to_path.setdefault(uid, path)
            self.log_tensor(f"{path}::{args[1]}", args[0])
        if self.verbose:
            print(f"{'[BW]' if self.mod_tracker.is_bw else '[FW]'} `{path}`: {func}")
        return out


def _find_all_submodules_compiled(out: List[nn.Module], module: nn.Module) -> None:
    if module._compiled_call_impl is not None:
        out.append(module)
    for c in module.children():
        _find_all_submodules_compiled(out, module=c)


class TorchCompileDisabler:
    def __init__(self, module: nn.Module) -> None:
        self.module = module
        self.submodules_compiled: List[nn.Module] = []
        self.compiled_call_impl: List[Any] = []
        self.disable_compile = torch.compiler.disable()
        torch._dynamo.config.raise_on_ctx_manager_usage = False  # type: ignore

    def __enter__(self) -> None:
        # Remove all `_compiled_call_impl` attributes to effectively
        # "undo" compilation
        self.submodules_compiled.clear()
        _find_all_submodules_compiled(self.submodules_compiled, self.module)
        self.compiled_call_impl = [
            m._compiled_call_impl for m in self.submodules_compiled
        ]
        for m in self.submodules_compiled:
            m._compiled_call_impl = None
        self.disable_compile.__enter__()  # type: ignore

    def __exit__(self, *args) -> None:
        self.disable_compile.__exit__(*args)  # type: ignore
        for m, c_impl in zip(self.submodules_compiled, self.compiled_call_impl):
            m._compiled_call_impl = c_impl
        self.compiled_call_impl = []


Probe = AutoProbeD

# EXAMPLE USAGE
d = 512
seqlen = 4
bs = 2


class Attention1(nn.Module):
    def forward(self, x):
        attn_bias = fmha.attn_bias.LowerTriangularFromBottomRightMask()
        return fmha.memory_efficient_attention(x, x, x, attn_bias=attn_bias).reshape(
            [x.shape[0], seqlen, -1]
        )


class Attention2(nn.Module):
    def forward(self, x):
        attn_bias = fmha.attn_bias.BlockDiagonalMask.from_seqlens(
            [seqlen] * bs
        ).make_causal()
        xr = x.reshape([1, 2 * seqlen, x.shape[2], x.shape[3]])
        return fmha.memory_efficient_attention(xr, xr, xr, attn_bias=attn_bias).reshape(
            [x.shape[0], seqlen, -1]
        )


class AttentionSDPA(nn.Module):
    def __init__(self):
        super().__init__()
        self.wo = nn.Linear(d, d)

    def forward(self, x):
        x = x.transpose(1, 2)
        return self.wo(
            F.scaled_dot_product_attention(x, x, x)
            .transpose(1, 2)
            .reshape([x.shape[0], seqlen, -1])
        )


class AttentionSDPAFlash(AttentionSDPA):
    def forward(self, x):
        x = x.transpose(1, 2)
        with sdpa_kernel(SDPBackend.FLASH_ATTENTION):
            return self.wo(
                F.scaled_dot_product_attention(x, x, x)
                .transpose(1, 2)
                .reshape([x.shape[0], seqlen, -1])
            )


class Model(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.head = nn.Linear(d, 16)
        self.trunk = nn.Sequential(
            nn.Linear(d, d),
            nn.Linear(d, d),
        )
        self.q_proj = nn.Linear(d, d, bias=False)
        self.trunk.compile()
        self.attn1 = Attention1()
        self.attn2 = Attention2()
        self.attnSDPA = AttentionSDPA()
        self.attnSDPAflash = AttentionSDPAFlash()

    def forward(self, x):
        B, nHeads, D = x.shape[0], d // 64, 64
        x = self.q_proj(x).reshape([B, seqlen, nHeads, D])
        x = self.attn1(x) + self.attn2(x) + self.attnSDPA(x) + self.attnSDPAflash(x)
        x = log_stats(x, "attns_out")
        return self.head(self.trunk(x))


def test_masking() -> None:
    q_seqlen = [1, 1, 14, 12]
    kv_seqlen = [2, 2, 14, 18]
    attn_bias = fmha.attn_bias.BlockDiagonalCausalMask.from_seqlens(
        q_seqlen, kv_seqlen
    ).make_causal_from_bottomright()
    logits = torch.randn(
        [1, 1, sum(q_seqlen), sum(kv_seqlen)], dtype=torch.float32, device="cuda"
    )
    bias = attn_bias.materialize(logits.shape, dtype=logits.dtype, device=logits.device)
    logits_masked = logits.clone()
    _mask_attn_logits(
        logits_masked,
        list(range(logits.shape[2])),
        causal=True,
        cu_seqlens_q=attn_bias.q_seqinfo.seqstart,
        cu_seqlens_k=attn_bias.k_seqinfo.seqstart,
    )
    assert (logits + bias == logits_masked).all().item()


def test_toy_model() -> None:
    # Test masking
    kw = dict(device="cuda", dtype=torch.float16)
    x = torch.randn([bs, seqlen, d], **kw)
    m = Model()
    m.head = checkpoint_wrapper(
        m.head, checkpoint_impl=CheckpointImpl.NO_REENTRANT, preserve_rng_state=False
    )
    m.to(**kw)
    m.compile()
    optim = torch.optim.SGD(m.parameters(), lr=0.0)
    probe = AutoProbeD(m, "./probe.json")

    for i in range(4):
        with contextlib.ExitStack() as stack:
            print(f"########### STEP {i}")
            if i % 4 == 1:
                stack.enter_context(probe)
                probe.metadata = {"it": i}
            y = m(x)
            g = torch.randn_like(y)
            y.backward(g)
            if i % 4 == 1:
                assert probe.enabled
                # Make sure we registered all linears
                print(list(probe.store.keys()))
                for key in [
                    "Model::attns_out",
                    "Model::attns_out.g",
                    "Model.attn1::attn_logits",
                    "Model.attn2::attn_logits",
                    "Model.attnSDPA::attn_logits",
                    "Model.attnSDPAflash::attn_logits",
                    "Model.head::w",
                    "Model.head::w.g",
                    "Model.head::in",
                    "Model.head::in.g",
                    "Model.head::out",
                    "Model.head::out.g",
                    "Model.trunk.0::in",
                    "Model.trunk.1::in",
                ]:
                    assert key in probe.store, f"Missing key: '{key}'"
                # .. and that the values are correct
                for key, tensor in [
                    ("Model.head::w", m.head.weight),
                    ("Model.head::w.g", m.head.weight.grad),
                    ("Model.q_proj::in", x),
                    ("Model.q_proj::w.g", m.q_proj.weight.grad),
                    ("Model.head::out", y),
                    ("Model.head::out.g", g),
                ]:
                    assert key in probe.store, f"Missing key: '{key}'"
                    assert torch.allclose(
                        probe.store[key]["abs.mean"], tensor.float().abs().mean()
                    ), f"'{key}' mismatches"
                # Check we don't have `nans`
                for key, value in probe.store.items():
                    if "abs.mean" in value:
                        assert math.isfinite(
                            value["abs.mean"].item()
                        ), f"Inf/Nan for {key}"
            optim.step()
            optim.zero_grad()



"""Audio-Visual Synchronisation Detection module.

Measures the temporal offset between audio and video streams.

Algorithm:
1. Extract per-frame visual activity (mean absolute frame
   difference in luminance → "visual energy" signal).
2. Extract per-frame audio energy (RMS of audio samples
   corresponding to each frame interval).
3. Cross-correlate the two energy signals.
4. The lag at maximum correlation is the estimated A/V offset.

Result is in milliseconds — positive means audio leads video.
|offset| < 40 ms is generally imperceptible; > 100 ms is noticeable.

Requires ffmpeg for audio extraction.
"""

import logging
import subprocess
import tempfile
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class AudioVisualSyncModule(PipelineModule):
    name = "av_sync"
    description = "Audio-video synchronisation offset detection"
    default_config = {
        "max_frames": 600,
        "warning_threshold_ms": 80.0,  # Warn if |offset| > 80 ms
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.max_frames = self.config.get("max_frames", 600)
        self.warning_threshold = self.config.get("warning_threshold_ms", 80.0)

    # ------------------------------------------------------------------
    def _extract_audio_pcm(self, video_path: Path, sr: int = 16000) -> Optional[np.ndarray]:
        """Extract raw mono PCM from video via ffmpeg."""
        try:
            tmp = tempfile.NamedTemporaryFile(suffix=".wav", delete=False)
            tmp.close()

            cmd = [
                "ffmpeg", "-y", "-i", str(video_path),
                "-vn", "-ac", "1", "-ar", str(sr),
                "-sample_fmt", "s16", tmp.name,
            ]
            result = subprocess.run(cmd, capture_output=True, timeout=30)
            if result.returncode != 0:
                Path(tmp.name).unlink(missing_ok=True)
                return None

            import soundfile as sf

            data, _ = sf.read(tmp.name, dtype="float32")
            Path(tmp.name).unlink(missing_ok=True)

            if data.ndim > 1:
                data = data.mean(axis=1)
            return data

        except ImportError:
            logger.warning("soundfile not installed")
            return None
        except FileNotFoundError:
            logger.warning("ffmpeg not found")
            return None
        except Exception as e:
            logger.debug(f"Audio extraction failed: {e}")
            return None

    def _audio_energy_per_frame(
        self, audio: np.ndarray, sr: int, fps: float, n_frames: int
    ) -> np.ndarray:
        """Compute RMS energy of audio aligned to each video frame."""
        samples_per_frame = int(sr / fps)
        energies = []
        for i in range(n_frames):
            start = i * samples_per_frame
            end = start + samples_per_frame
            if end > len(audio):
                break
            chunk = audio[start:end]
            rms = float(np.sqrt(np.mean(chunk ** 2)))
            energies.append(rms)
        return np.array(energies, dtype=np.float64)

    # ------------------------------------------------------------------
    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample

        # Check that the video has an audio stream
        if sample.audio_metadata is None:
            return sample

        cap = cv2.VideoCapture(str(sample.path))
        if not cap.isOpened():
            return sample

        fps = cap.get(cv2.CAP_PROP_FPS)
        if fps <= 0:
            cap.release()
            return sample

        try:
            # --- visual energy ---
            prev_gray = None
            visual_energy = []
            idx = 0

            while idx < self.max_frames:
                ret, frame = cap.read()
                if not ret:
                    break
                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)

                if prev_gray is not None:
                    diff = np.abs(gray - prev_gray)
                    visual_energy.append(float(diff.mean()))
                else:
                    visual_energy.append(0.0)

                prev_gray = gray
                idx += 1

            cap.release()

            if len(visual_energy) < 10:
                return sample

            # --- audio energy ---
            sr = 16000
            audio = self._extract_audio_pcm(sample.path, sr=sr)
            if audio is None or len(audio) < sr * 0.5:
                return sample

            n_frames = len(visual_energy)
            audio_energy = self._audio_energy_per_frame(audio, sr, fps, n_frames)

            # Align lengths
            min_len = min(len(visual_energy), len(audio_energy))
            if min_len < 10:
                return sample

            v_sig = np.array(visual_energy[:min_len])
            a_sig = audio_energy[:min_len]

            # --- cross-correlation ---
            # Normalise both signals to zero-mean, unit-variance
            v_sig = v_sig - v_sig.mean()
            a_sig = a_sig - a_sig.mean()

            v_std = v_sig.std()
            a_std = a_sig.std()

            if v_std < 1e-8 or a_std < 1e-8:
                # One signal is essentially flat — can't estimate offset
                return sample

            v_sig /= v_std
            a_sig /= a_std

            corr = np.correlate(a_sig, v_sig, mode="full")
            corr /= min_len  # normalise

            # Lag axis: negative lag means audio leads
            lags = np.arange(-(min_len - 1), min_len)
            best_lag_idx = int(np.argmax(corr))
            best_lag_frames = lags[best_lag_idx]

            offset_ms = float(best_lag_frames / fps * 1000.0)

            # Store result
            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.av_sync_offset = offset_ms

            if abs(offset_ms) > self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"A/V sync offset: {offset_ms:+.1f} ms",
                        details={
                            "offset_ms": offset_ms,
                            "threshold_ms": self.warning_threshold,
                        },
                        recommendation=(
                            "Audio and video are noticeably out of sync. "
                            "Check muxing, frame rate conversion, or "
                            "audio processing pipeline."
                        ),
                    )
                )

            logger.debug(
                f"A/V sync for {sample.path.name}: {offset_ms:+.1f} ms "
                f"(lag={best_lag_frames} frames)"
            )

        except Exception as e:
            logger.error(f"A/V sync detection failed for {sample.path}: {e}")
            if cap.isOpened():
                cap.release()

        return sample


class AudioVisualSyncCompatModule(AudioVisualSyncModule):
    """Compatibility alias matching filename-based discovery."""

    name = "audio_visual_sync"



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import io
import json
import logging
import os
import pickle
import re
import shutil
import urllib
import urllib.error
import urllib.request
from typing import Optional
from urllib.parse import urlparse

import numpy as np
import pandas as pd
import yaml
from iopath.common.download import download
from iopath.common.file_io import file_lock, g_pathmgr
from .registry import registry
from torch.utils.model_zoo import tqdm
from torchvision.datasets.utils import (
    check_integrity,
    download_file_from_google_drive,
    extract_archive,
)


def now():
    from datetime import datetime

    return datetime.now().strftime("%Y%m%d%H%M")[:-1]


def is_url(url_or_filename):
    parsed = urlparse(url_or_filename)
    return parsed.scheme in ("http", "https")


def get_cache_path(rel_path):
    return os.path.expanduser(os.path.join(registry.get_path("cache_root"), rel_path))


def get_abs_path(rel_path):
    return os.path.join(registry.get_path("library_root"), rel_path)


def load_json(filename):
    with open(filename, "r") as f:
        return json.load(f)


# The following are adapted from torchvision and vissl
# torchvision: https://github.com/pytorch/vision
# vissl: https://github.com/facebookresearch/vissl/blob/main/vissl/utils/download.py


def makedir(dir_path):
    """
    Create the directory if it does not exist.
    """
    is_success = False
    try:
        if not g_pathmgr.exists(dir_path):
            g_pathmgr.mkdirs(dir_path)
        is_success = True
    except BaseException:
        print(f"Error creating directory: {dir_path}")
    return is_success


def get_redirected_url(url: str):
    """
    Given a URL, returns the URL it redirects to or the
    original URL in case of no indirection
    """
    import requests

    with requests.Session() as session:
        with session.get(url, stream=True, allow_redirects=True) as response:
            if response.history:
                return response.url
            else:
                return url


def to_google_drive_download_url(view_url: str) -> str:
    """
    Utility function to transform a view URL of google drive
    to a download URL for google drive
    Example input:
        https://drive.google.com/file/d/137RyRjvTBkBiIfeYBNZBtViDHQ6_Ewsp/view
    Example output:
        https://drive.google.com/uc?export=download&id=137RyRjvTBkBiIfeYBNZBtViDHQ6_Ewsp
    """
    splits = view_url.split("/")
    assert splits[-1] == "view"
    file_id = splits[-2]
    return f"https://drive.google.com/uc?export=download&id={file_id}"


def download_google_drive_url(url: str, output_path: str, output_file_name: str):
    """
    Download a file from google drive
    Downloading an URL from google drive requires confirmation when
    the file of the size is too big (google drive notifies that
    anti-viral checks cannot be performed on such files)
    """
    import requests

    with requests.Session() as session:

        # First get the confirmation token and append it to the URL
        with session.get(url, stream=True, allow_redirects=True) as response:
            for k, v in response.cookies.items():
                if k.startswith("download_warning"):
                    url = url + "&confirm=" + v

        # Then download the content of the file
        with session.get(url, stream=True, verify=True) as response:
            makedir(output_path)
            path = os.path.join(output_path, output_file_name)
            total_size = int(response.headers.get("Content-length", 0))
            with open(path, "wb") as file:
                from tqdm import tqdm

                with tqdm(total=total_size) as progress_bar:
                    for block in response.iter_content(
                        chunk_size=io.DEFAULT_BUFFER_SIZE
                    ):
                        file.write(block)
                        progress_bar.update(len(block))


def _get_google_drive_file_id(url: str) -> Optional[str]:
    parts = urlparse(url)

    if re.match(r"(drive|docs)[.]google[.]com", parts.netloc) is None:
        return None

    match = re.match(r"/file/d/(?P<id>[^/]*)", parts.path)
    if match is None:
        return None

    return match.group("id")


def _urlretrieve(url: str, filename: str, chunk_size: int = 1024) -> None:
    with open(filename, "wb") as fh:
        with urllib.request.urlopen(
            urllib.request.Request(url, headers={"User-Agent": "vissl"})
        ) as response:
            with tqdm(total=response.length) as pbar:
                for chunk in iter(lambda: response.read(chunk_size), ""):
                    if not chunk:
                        break
                    pbar.update(chunk_size)
                    fh.write(chunk)


def download_url(
    url: str,
    root: str,
    filename: Optional[str] = None,
    md5: Optional[str] = None,
) -> None:
    """Download a file from a url and place it in root.
    Args:
        url (str): URL to download file from
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under.
                                  If None, use the basename of the URL.
        md5 (str, optional): MD5 checksum of the download. If None, do not check
    """
    root = os.path.expanduser(root)
    if not filename:
        filename = os.path.basename(url)
    fpath = os.path.join(root, filename)

    makedir(root)

    # check if file is already present locally
    if check_integrity(fpath, md5):
        print("Using downloaded and verified file: " + fpath)
        return

    # expand redirect chain if needed
    url = get_redirected_url(url)

    # check if file is located on Google Drive
    file_id = _get_google_drive_file_id(url)
    if file_id is not None:
        return download_file_from_google_drive(file_id, root, filename, md5)

    # download the file
    try:
        print("Downloading " + url + " to " + fpath)
        _urlretrieve(url, fpath)
    except (urllib.error.URLError, IOError) as e:  # type: ignore[attr-defined]
        if url[:5] == "https":
            url = url.replace("https:", "http:")
            print(
                "Failed download. Trying https -> http instead."
                " Downloading " + url + " to " + fpath
            )
            _urlretrieve(url, fpath)
        else:
            raise e

    # check integrity of downloaded file
    if not check_integrity(fpath, md5):
        raise RuntimeError("File not found or corrupted.")


def download_and_extract_archive(
    url: str,
    download_root: str,
    extract_root: Optional[str] = None,
    filename: Optional[str] = None,
    md5: Optional[str] = None,
    remove_finished: bool = False,
) -> None:
    download_root = os.path.expanduser(download_root)
    if extract_root is None:
        extract_root = download_root
    if not filename:
        filename = os.path.basename(url)

    download_url(url, download_root, filename, md5)

    archive = os.path.join(download_root, filename)
    print("Extracting {} to {}".format(archive, extract_root))
    extract_archive(archive, extract_root, remove_finished)


def cache_url(url: str, cache_dir: str) -> str:
    """
    This implementation downloads the remote resource and caches it locally.
    The resource will only be downloaded if not previously requested.
    """
    parsed_url = urlparse(url)
    dirname = os.path.join(cache_dir, os.path.dirname(parsed_url.path.lstrip("/")))
    makedir(dirname)
    filename = url.split("/")[-1]
    cached = os.path.join(dirname, filename)
    with file_lock(cached):
        if not os.path.isfile(cached):
            logging.info(f"Downloading {url} to {cached} ...")
            cached = download(url, dirname, filename=filename)
    logging.info(f"URL {url} cached in {cached}")
    return cached


# TODO (prigoyal): convert this into RAII-style API
def create_file_symlink(file1, file2):
    """
    Simply create the symlinks for a given file1 to file2.
    Useful during model checkpointing to symlinks to the
    latest successful checkpoint.
    """
    try:
        if g_pathmgr.exists(file2):
            g_pathmgr.rm(file2)
        g_pathmgr.symlink(file1, file2)
    except Exception as e:
        logging.info(f"Could NOT create symlink. Error: {e}")


def save_file(data, filename, append_to_json=True, verbose=True):
    """
    Common i/o utility to handle saving data to various file formats.
    Supported:
        .pkl, .pickle, .npy, .json
    Specifically for .json, users have the option to either append (default)
    or rewrite by passing in Boolean value to append_to_json.
    """
    if verbose:
        logging.info(f"Saving data to file: {filename}")
    file_ext = os.path.splitext(filename)[1]
    if file_ext in [".pkl", ".pickle"]:
        with g_pathmgr.open(filename, "wb") as fopen:
            pickle.dump(data, fopen, pickle.HIGHEST_PROTOCOL)
    elif file_ext == ".npy":
        with g_pathmgr.open(filename, "wb") as fopen:
            np.save(fopen, data)
    elif file_ext == ".json":
        if append_to_json:
            with g_pathmgr.open(filename, "a") as fopen:
                fopen.write(json.dumps(data, sort_keys=True) + "\n")
                fopen.flush()
        else:
            with g_pathmgr.open(filename, "w") as fopen:
                fopen.write(json.dumps(data, sort_keys=True) + "\n")
                fopen.flush()
    elif file_ext == ".yaml":
        with g_pathmgr.open(filename, "w") as fopen:
            dump = yaml.dump(data)
            fopen.write(dump)
            fopen.flush()
    else:
        raise Exception(f"Saving {file_ext} is not supported yet")

    if verbose:
        logging.info(f"Saved data to file: {filename}")


def load_file(filename, mmap_mode=None, verbose=True, allow_pickle=False):
    """
    Common i/o utility to handle loading data from various file formats.
    Supported:
        .pkl, .pickle, .npy, .json
    For the npy files, we support reading the files in mmap_mode.
    If the mmap_mode of reading is not successful, we load data without the
    mmap_mode.
    """
    if verbose:
        logging.info(f"Loading data from file: {filename}")

    file_ext = os.path.splitext(filename)[1]
    if file_ext == ".txt":
        with g_pathmgr.open(filename, "r") as fopen:
            data = fopen.readlines()
    elif file_ext in [".pkl", ".pickle"]:
        with g_pathmgr.open(filename, "rb") as fopen:
            data = pickle.load(fopen, encoding="latin1")
    elif file_ext == ".npy":
        if mmap_mode:
            try:
                with g_pathmgr.open(filename, "rb") as fopen:
                    data = np.load(
                        fopen,
                        allow_pickle=allow_pickle,
                        encoding="latin1",
                        mmap_mode=mmap_mode,
                    )
            except ValueError as e:
                logging.info(
                    f"Could not mmap {filename}: {e}. Trying without g_pathmgr"
                )
                data = np.load(
                    filename,
                    allow_pickle=allow_pickle,
                    encoding="latin1",
                    mmap_mode=mmap_mode,
                )
                logging.info("Successfully loaded without g_pathmgr")
            except Exception:
                logging.info("Could not mmap without g_pathmgr. Trying without mmap")
                with g_pathmgr.open(filename, "rb") as fopen:
                    data = np.load(fopen, allow_pickle=allow_pickle, encoding="latin1")
        else:
            with g_pathmgr.open(filename, "rb") as fopen:
                data = np.load(fopen, allow_pickle=allow_pickle, encoding="latin1")
    elif file_ext == ".json":
        with g_pathmgr.open(filename, "r") as fopen:
            data = json.load(fopen)
    elif file_ext == ".yaml":
        with g_pathmgr.open(filename, "r") as fopen:
            data = yaml.load(fopen, Loader=yaml.FullLoader)
    elif file_ext == ".csv":
        with g_pathmgr.open(filename, "r") as fopen:
            data = pd.read_csv(fopen)
    else:
        raise Exception(f"Reading from {file_ext} is not supported yet")
    return data


def abspath(resource_path: str):
    """
    Make a path absolute, but take into account prefixes like
    "http://" or "manifold://"
    """
    regex = re.compile(r"^\w+://")
    if regex.match(resource_path) is None:
        return os.path.abspath(resource_path)
    else:
        return resource_path


def makedir(dir_path):
    """
    Create the directory if it does not exist.
    """
    is_success = False
    try:
        if not g_pathmgr.exists(dir_path):
            g_pathmgr.mkdirs(dir_path)
        is_success = True
    except BaseException:
        logging.info(f"Error creating directory: {dir_path}")
    return is_success


def is_url(input_url):
    """
    Check if an input string is a url. look for http(s):// and ignoring the case
    """
    is_url = re.match(r"^(?:http)s?://", input_url, re.IGNORECASE) is not None
    return is_url


def cleanup_dir(dir):
    """
    Utility for deleting a directory. Useful for cleaning the storage space
    that contains various training artifacts like checkpoints, data etc.
    """
    if os.path.exists(dir):
        logging.info(f"Deleting directory: {dir}")
        shutil.rmtree(dir)
    logging.info(f"Deleted contents of directory: {dir}")


def get_file_size(filename):
    """
    Given a file, get the size of file in MB
    """
    size_in_mb = os.path.getsize(filename) / float(1024**2)
    return size_in_mb



"""UTMOS (Universal Text-to-Speech MOS) module.

No-reference MOS prediction for speech quality using a neural model.
Predicts the Mean Opinion Score that human listeners would assign,
without needing a reference audio.

Score range: 1.0 to 5.0 (higher = better, MOS scale).
  4.0+  excellent (natural speech)
  3.5+  good
  3.0+  fair
  <3.0  poor

Uses the SpeechMOS model (sarulab-speech/speechmos) from Hugging Face.

References:
    - Saeki et al. (2022), "UTMOS: UTokyo-SaruLab System for
      VoiceMOS Challenge 2022"
"""

import logging
from pathlib import Path
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class AudioUTMOSModule(PipelineModule):
    name = "audio_utmos"
    description = "UTMOS no-reference MOS prediction for speech quality"
    default_config = {
        "target_sr": 16000,
        "warning_threshold": 3.0,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.target_sr = self.config.get("target_sr", 16000)
        self.warning_threshold = self.config.get("warning_threshold", 3.0)
        self._model = None
        self._ml_available = False

    def setup(self) -> None:
        try:
            import torch

            self._device = "cuda" if torch.cuda.is_available() else "cpu"
            logger.info(f"Loading UTMOS model on {self._device}...")
            self._model = torch.hub.load(
                "tarepan/SpeechMOS:v1.2.0", "utmos22_strong",
                trust_repo=True,
            )
            self._model = self._model.to(self._device)
            self._model.eval()
            self._ml_available = True
            logger.info("UTMOS module initialised")

        except ImportError:
            logger.warning("torch not installed. UTMOS requires PyTorch.")
        except Exception as e:
            logger.warning(f"Failed to setup UTMOS: {e}")

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available:
            return sample

        try:
            audio = self._load_audio(sample.path)
            if audio is None:
                # For video files, try extracting audio via ffmpeg
                if sample.is_video:
                    audio = self._extract_audio_from_video(sample.path)
                if audio is None:
                    return sample

            import torch

            waveform = torch.from_numpy(audio).unsqueeze(0).to(self._device)

            with torch.no_grad():
                score = self._model(waveform, self.target_sr).item()

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.utmos_score = round(float(score), 3)

            if score < self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low predicted MOS (UTMOS): {score:.2f}",
                        details={"utmos": float(score)},
                    )
                )

        except Exception as e:
            logger.warning(f"UTMOS failed for {sample.path}: {e}")

        return sample

    def _extract_audio_from_video(self, path: Path) -> Optional[np.ndarray]:
        """Extract audio from a video file via ffmpeg, return as float32 array."""
        import subprocess
        import tempfile

        try:
            tmp = tempfile.NamedTemporaryFile(suffix=".wav", delete=False)
            tmp.close()
            cmd = [
                "ffmpeg", "-y", "-i", str(path),
                "-vn", "-ac", "1", "-ar", str(self.target_sr),
                "-sample_fmt", "s16", tmp.name,
            ]
            result = subprocess.run(cmd, capture_output=True, timeout=30)
            if result.returncode != 0:
                Path(tmp.name).unlink(missing_ok=True)
                return None
            audio = self._load_audio(Path(tmp.name))
            Path(tmp.name).unlink(missing_ok=True)
            return audio
        except FileNotFoundError:
            logger.debug("ffmpeg not found — cannot extract audio from video")
            return None
        except Exception as e:
            logger.debug(f"Video audio extraction failed: {e}")
            return None

    def _load_audio(self, path: Path) -> Optional[np.ndarray]:
        try:
            import soundfile as sf

            data, sr = sf.read(str(path), dtype="float32")
            if sr != self.target_sr:
                import librosa

                data = librosa.resample(data, orig_sr=sr, target_sr=self.target_sr)
            if data.ndim > 1:
                data = data.mean(axis=1)
            return data.astype(np.float32)
        except ImportError:
            logger.warning("soundfile/librosa not installed")
            return None
        except Exception as e:
            logger.debug(f"Audio load error: {e}")
            return None



import logging
import os
import json
import sqlite3
import random
from os.path import basename

import numpy as np

from .base_dataset import ImageVideoBaseDataset
from .utils import load_anno, pre_text
from .video_utils import VIDEO_READER_FUNCS
from ..utils.distributed import is_main_process

logger = logging.getLogger(__name__)


def get_anno_by_id(cur: sqlite3.Cursor, id: int):
    """TODO: Docstring for get_anno_by_id.

    Args:
        cur (sqlite3.Cursor): The dataset cursor.
        id (int): The annotation id.

    Returns:

    """
    pass


class SQLiteImgTxtRetTrainDataset(ImageVideoBaseDataset):
    media_type = "image"

    def __init__(self, ann_file, transform, has_multi_vision_gt=False):
        super().__init__()

        if len(ann_file) == 3 and ann_file[2] == "video":
            self.media_type = "video"  
        else:
            self.media_type = "image"
        self.label_file, self.data_root = ann_file[:2]

        if '.json' in self.label_file:
            logger.info('Load json file')
            with open(self.label_file, 'r') as f:
                self.anno = json.load(f)
            self.num_examples = len(self.anno)
        else:
            logger.info('Load sql file')
            self.con = sqlite3.connect("file:" + self.label_file + "?mode=ro", uri=True)
            self.cur = self.con.cursor()
            self.num_examples = self.cur.execute("SELECT COUNT(*) FROM annos").fetchone()[0]

        self.transform = transform
        # each caption has multiple image as ground_truth, e.g., ssv2
        self.has_multi_vision_gt = has_multi_vision_gt
        assert not self.has_multi_vision_gt

    def get_anno(self, index):
        if '.json' in self.label_file:
            filename = self.anno[index][self.media_type]
            caption = self.anno[index]["caption"]
        else:
            query = f"SELECT * FROM annos WHERE id = {index};"
            res = self.cur.execute(query)
            id, filename, caption = res.fetchone()
        anno = {"image": os.path.join(self.data_root, filename), "caption": caption}
        return anno

    def __len__(self):
        return self.num_examples

    def __getitem__(self, index):
        try:
            ann = self.get_anno(index)
            image, index = self.load_and_transform_media_data(index, ann["image"])
            caption = pre_text(ann["caption"])
            # key = ann["caption"] if self.has_multi_vision_gt else basename(ann["image"])
            return image, caption, index
        except Exception as e:
            logger.warning(f"Caught exception {e} when loading image {ann['image']}")
            index = np.random.randint(0, len(self))
            return self.__getitem__(index)


class SQLiteVidTxtRetTrainDataset(SQLiteImgTxtRetTrainDataset):
    media_type = "video"

    def __init__(
        self,
        ann_file,
        transform,
        num_frames=4,
        video_reader_type="decord",
        sample_type="rand",
        num_tries=3,
        is_paragraph_retrieval=False,
        has_multi_vision_gt=False,
        repeat_kinetics=1,
    ):
        super().__init__(ann_file, transform, has_multi_vision_gt)
        self.num_frames = num_frames
        self.video_reader_type = video_reader_type
        self.video_reader = VIDEO_READER_FUNCS[video_reader_type]
        self.sample_type = sample_type
        self.num_tries = num_tries
        self.is_paragraph_retrieval = is_paragraph_retrieval

        if is_paragraph_retrieval:
            raise ValueError(f"not implemented")


from math import log, pi
from typing import Literal, Optional, Union

import torch
from einops import rearrange, repeat
from torch import Tensor, broadcast_tensors, einsum, nn
from torch.amp import autocast
from torch.nn import Module, ModuleList

# helper functions


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


# broadcat, as tortoise-tts was using it


def broadcat(tensors, dim=-1):
    broadcasted_tensors = broadcast_tensors(*tensors)
    return torch.cat(broadcasted_tensors, dim=dim)


# rotary embedding helper functions


def rotate_half(x):
    x = rearrange(x, "... (d r) -> ... d r", r=2)
    x1, x2 = x.unbind(dim=-1)
    x = torch.stack((-x2, x1), dim=-1)
    return rearrange(x, "... d r -> ... (d r)")


@autocast("cuda", enabled=False)
def apply_rotary_emb(freqs, t, start_index=0, scale=1.0, seq_dim=-2):
    dtype = t.dtype

    if t.ndim == 3:
        seq_len = t.shape[seq_dim]
        freqs = freqs[-seq_len:]

    rot_dim = freqs.shape[-1]
    end_index = start_index + rot_dim

    assert (
        rot_dim <= t.shape[-1]
    ), f"feature dimension {t.shape[-1]} is not of sufficient size to rotate in all the positions {rot_dim}"

    t_left, t, t_right = (
        t[..., :start_index],
        t[..., start_index:end_index],
        t[..., end_index:],
    )
    t = (t * freqs.cos() * scale) + (rotate_half(t) * freqs.sin() * scale)
    out = torch.cat((t_left, t, t_right), dim=-1)

    return out.type(dtype)


# learned rotation helpers


def apply_learned_rotations(rotations, t, start_index=0, freq_ranges=None):
    if exists(freq_ranges):
        rotations = einsum("..., f -> ... f", rotations, freq_ranges)
        rotations = rearrange(rotations, "... r f -> ... (r f)")

    rotations = repeat(rotations, "... n -> ... (n r)", r=2)
    return apply_rotary_emb(rotations, t, start_index=start_index)


# classes


class RotaryEmbedding(Module):
    def __init__(
        self,
        dim,
        custom_freqs: Optional[Tensor] = None,
        freqs_for: Union[
            Literal["lang"], Literal["pixel"], Literal["constant"]
        ] = "lang",
        theta=10000,
        max_freq=10,
        num_freqs=1,
        learned_freq=False,
        use_xpos=False,
        xpos_scale_base=512,
        interpolate_factor=1.0,
        theta_rescale_factor=1.0,
        seq_before_head_dim=False,
        cache_if_possible=True,
    ):
        super().__init__()
        # proposed by reddit user bloc97, to rescale rotary embeddings to longer sequence length without fine-tuning
        # has some connection to NTK literature
        # https://www.reddit.com/r/LocalLLaMA/comments/14lz7j5/ntkaware_scaled_rope_allows_llama_models_to_have/

        theta *= theta_rescale_factor ** (dim / (dim - 2))

        self.freqs_for = freqs_for

        if exists(custom_freqs):
            freqs = custom_freqs
        elif freqs_for == "lang":
            freqs = 1.0 / (
                theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim)
            )
        elif freqs_for == "pixel":
            freqs = torch.linspace(1.0, max_freq / 2, dim // 2) * pi
        elif freqs_for == "constant":
            freqs = torch.ones(num_freqs).float()

        self.cache_if_possible = cache_if_possible

        self.tmp_store("cached_freqs", None)
        self.tmp_store("cached_scales", None)

        self.freqs = nn.Parameter(freqs, requires_grad=learned_freq)

        self.learned_freq = learned_freq

        # dummy for device

        self.tmp_store("dummy", torch.tensor(0))

        # default sequence dimension

        self.seq_before_head_dim = seq_before_head_dim
        self.default_seq_dim = -3 if seq_before_head_dim else -2

        # interpolation factors

        assert interpolate_factor >= 1.0
        self.interpolate_factor = interpolate_factor

        # xpos

        self.use_xpos = use_xpos
        if not use_xpos:
            self.tmp_store("scale", None)
            return

        scale = (torch.arange(0, dim, 2) + 0.4 * dim) / (1.4 * dim)

        self.scale_base = xpos_scale_base
        self.tmp_store("scale", scale)

        # add apply_rotary_emb as static method

        self.apply_rotary_emb = staticmethod(apply_rotary_emb)

    @property
    def device(self):
        return self.dummy.device

    def tmp_store(self, key, value):
        self.register_buffer(key, value, persistent=False)

    def get_seq_pos(self, seq_len, device, dtype, offset=0):
        return (
            torch.arange(seq_len, device=device, dtype=dtype) + offset
        ) / self.interpolate_factor

    def rotate_queries_or_keys(self, t, seq_dim=None, offset=0):
        seq_dim = default(seq_dim, self.default_seq_dim)

        assert (
            not self.use_xpos
        ), "you must use `.rotate_queries_and_keys` method instead and pass in both queries and keys, for length extrapolatable rotary embeddings"

        device, dtype, seq_len = t.device, t.dtype, t.shape[seq_dim]

        freqs = self.forward(
            self.get_seq_pos(seq_len, device=device, dtype=dtype, offset=offset),
            seq_len=seq_len,
            offset=offset,
        )

        if seq_dim == -3:
            freqs = rearrange(freqs, "n d -> n 1 d")

        return apply_rotary_emb(freqs, t, seq_dim=seq_dim)

    def rotate_queries_with_cached_keys(self, q, k, seq_dim=None, offset=0):
        seq_dim = default(seq_dim, self.default_seq_dim)

        q_len, k_len = q.shape[seq_dim], k.shape[seq_dim]
        assert q_len <= k_len

        rotated_q = self.rotate_queries_or_keys(
            q, seq_dim=seq_dim, offset=k_len - q_len + offset
        )
        rotated_k = self.rotate_queries_or_keys(k, seq_dim=seq_dim, offset=offset)

        rotated_q = rotated_q.type(q.dtype)
        rotated_k = rotated_k.type(k.dtype)

        return rotated_q, rotated_k

    def rotate_queries_and_keys(self, q, k, seq_dim=None):
        seq_dim = default(seq_dim, self.default_seq_dim)

        assert self.use_xpos
        device, dtype, seq_len = q.device, q.dtype, q.shape[seq_dim]

        seq = self.get_seq_pos(seq_len, dtype=dtype, device=device)

        freqs = self.forward(seq, seq_len=seq_len)
        scale = self.get_scale(seq, seq_len=seq_len).to(dtype)

        if seq_dim == -3:
            freqs = rearrange(freqs, "n d -> n 1 d")
            scale = rearrange(scale, "n d -> n 1 d")

        rotated_q = apply_rotary_emb(freqs, q, scale=scale, seq_dim=seq_dim)
        rotated_k = apply_rotary_emb(freqs, k, scale=scale**-1, seq_dim=seq_dim)

        rotated_q = rotated_q.type(q.dtype)
        rotated_k = rotated_k.type(k.dtype)

        return rotated_q, rotated_k

    def get_scale(self, t: Tensor, seq_len: Optional[int] = None, offset=0):
        assert self.use_xpos

        should_cache = self.cache_if_possible and exists(seq_len)

        if (
            should_cache
            and exists(self.cached_scales)
            and (seq_len + offset) <= self.cached_scales.shape[0]
        ):
            return self.cached_scales[offset : (offset + seq_len)]

        scale = 1.0
        if self.use_xpos:
            power = (t - len(t) // 2) / self.scale_base
            scale = self.scale ** rearrange(power, "n -> n 1")
            scale = torch.cat((scale, scale), dim=-1)

        if should_cache:
            self.tmp_store("cached_scales", scale)

        return scale

    def get_axial_freqs(self, *dims):
        Colon = slice(None)
        all_freqs = []

        for ind, dim in enumerate(dims):
            if self.freqs_for == "pixel":
                pos = torch.linspace(-1, 1, steps=dim, device=self.device)
            else:
                pos = torch.arange(dim, device=self.device)

            freqs = self.forward(pos, seq_len=dim)

            all_axis = [None] * len(dims)
            all_axis[ind] = Colon

            new_axis_slice = (Ellipsis, *all_axis, Colon)
            all_freqs.append(freqs[new_axis_slice])

        all_freqs = broadcast_tensors(*all_freqs)
        return torch.cat(all_freqs, dim=-1)

    @autocast("cuda", enabled=False)
    def forward(self, t: Tensor, seq_len=None, offset=0):
        should_cache = (
            self.cache_if_possible
            and not self.learned_freq
            and exists(seq_len)
            and self.freqs_for != "pixel"
        )

        if (
            should_cache
            and exists(self.cached_freqs)
            and (offset + seq_len) <= self.cached_freqs.shape[0]
        ):
            return self.cached_freqs[offset : (offset + seq_len)].detach()

        freqs = self.freqs

        freqs = einsum("..., f -> ... f", t.type(freqs.dtype), freqs)
        freqs = repeat(freqs, "... n -> ... (n r)", r=2)

        if should_cache:
            self.tmp_store("cached_freqs", freqs.detach())

        return freqs





class Rope2D:
    """ Helper class to apply RoPE2D as well as interpolate on the fly. """

    def __init__(self, dim, use_cls_token=False):
        self.dim = dim
        self.use_cls_token = use_cls_token
        self.grid_size = None
        self.freq = None

    def init_tensors(self):
        self.rope = RotaryEmbedding(self.dim // 2)

    def update_grid(self, device, grid_h, grid_w):
        if self.grid_size != (grid_h, grid_w):
            self.grid_size = (grid_h, grid_w)

            self.rope = self.rope.to(device)

            if self.use_cls_token:
                # +1 to leave space for the cls token to be (0, 0)
                grid_y_range = torch.arange(grid_h, device=device) + 1
                grid_x_range = torch.arange(grid_w, device=device) + 1
            else:
                grid_y_range = torch.arange(grid_h, device=device)
                grid_x_range = torch.arange(grid_w, device=device)

            freqs_y = self.rope(grid_y_range)[:, None].expand(grid_h, grid_w, -1)
            freqs_x = self.rope(grid_x_range)[None, :].expand(grid_h, grid_w, -1)
            freq = torch.cat([freqs_x, freqs_y], dim=-1).reshape(grid_h * grid_w, -1)

            if self.use_cls_token:
                freq = torch.cat(
                    [freq, torch.zeros(1, freq.shape[-1], device=device)], dim=0
                )

            self.freq = freq[None, ...]

        self.freq = self.freq.to(device)

    def __call__(self, q, k):
        # batch, heads, seq, dim = q.shape
        q = apply_rotary_emb(self.freq[:, None, :, :], q)
        k = apply_rotary_emb(self.freq[:, None, :, :], k)

        return q, k


import os
import torch
import numpy as np
import decord
from typing import List, Union
from transformers import AutoModel, AutoTokenizer
import torchvision.transforms as transforms
from torchvision.transforms import PILToTensor
from decord import VideoReader, cpu
import torch.nn.functional as F
from PIL import Image
from torchvision.transforms.functional import InterpolationMode
from typing import List, TypedDict, Union, Optional
decord.bridge.set_bridge("torch")

from .vqa_model import VQAScoreModel

INTERNVIDEO2_MODELS = {
    'internvideo2-chat-8b': {
        'tokenizer': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVideo2-Chat-8B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVideo2-Chat-8B',
            'torch_dtype': torch.bfloat16,
            'trust_remote_code': True,
        },
    },
    'internvideo2-chat-8b-hd': {
        'tokenizer': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVideo2_chat_8B_HD',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVideo2_chat_8B_HD',
            'torch_dtype': torch.bfloat16,
            'trust_remote_code': True,
        },
    },
    'internvideo2-chat-8b-internlm': {
        'tokenizer': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVideo2_Chat_8B_InternLM2_5',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVideo2_Chat_8B_InternLM2_5',
            'torch_dtype': torch.bfloat16,
            'trust_remote_code': True,
        },
    },
}

IMG_TOKEN = "[<IMG_PLH>]"
VID_TOKEN = "[<VID_PLH>]"

DEFAULT_PAD_TOKEN = "[PAD]"
DEFAULT_BOS_TOKEN = '<s>'
DEFAULT_EOS_TOKEN = '</s>'
DEFAULT_UNK_TOKEN = "<unk>"

DEFAULT_IMAGE_TOKEN = "[IMAGETOKEN]"
DEFAULT_VIDEO_TOKEN = "[VIDEOTOKEN]"

DEFAULT_IMG_PLACEHOLDER = "[<IMG_PLH>]"
DEFAULT_VID_PLACEHOLDER = "[<VID_PLH>]"
class InternVideo2Model(VQAScoreModel):
    video_mode = "direct"
    allows_image = False
    def __init__(self,
                 model_name='internvideo2-chat-8b',
                 device='cuda',
                 cache_dir=None):
        assert model_name in INTERNVIDEO2_MODELS, f"Model {model_name} not found in INTERNVIDEO2_MODELS"
        self.model_name = model_name
        self.device = device
        self.cache_dir = cache_dir
        self.model_info = INTERNVIDEO2_MODELS[model_name]
        self.is_hd = 'hd' in model_name
        self.load_model()

    def load_model(self):
        
        self.model = AutoModel.from_pretrained(
            **self.model_info['model']
        ).to(self.device)
        
        self.tokenizer = AutoTokenizer.from_pretrained(
            **self.model_info['tokenizer']
        )

    def load_video(self, video_path, num_segments=8, resolution=224, hd_num=6):
        if self.model_name == 'internvideo2-chat-8b':
            return self.load_video_chat(video_path, num_segments, return_msg=False)
        elif self.model_name == 'internvideo2-chat-8b-hd':
            return self.load_video_chat_hd(video_path, num_segments=8, return_msg=False, resolution=224, hd_num=6)

    def process_image(self, path, resolution=224):
        if self.model_name == 'internvideo2-chat-8b':
            return self.process_image_chat(path)
        elif self.model_name == 'internvideo2-chat-8b-hd':
            return self.process_image_chat_hd(path, resolution)
        

    def load_images(self, paths: List[str], num_segments: int = 8, resolution: int = 224, hd_num: int = 6) -> List[torch.Tensor]:
        processed_data = []
        for path in paths:
            if path.startswith(("http://", "https://")):
                raise NotImplementedError("Web link image/video inputs are not yet supported for this model. Please use a local path, or otherwise, make a Github issue request if this feature is necessary.")
            if path.lower().endswith(('.mp4', '.avi', '.mov', '.mkv')):  # Video file
                processed_data.append(self.load_video(path, num_segments=num_segments, resolution=resolution, hd_num=hd_num))
            else:  # Image file or .npy file
                print('This model does not yet support image inference.')
                exit()
                processed_data.append(self.process_image(path, resolution))
        return processed_data

    def forward(self,
                paths: List[str],
                texts: List[str],
                num_frames: int=8,
                question_template: str = "Does this figure show \"{}\"? Please answer yes or no.",
                answer_template: str = "Yes",
                num_segments: int = None,
                resolution: int = None,
                hd_num: int = None) -> torch.Tensor:
        assert len(paths) == len(texts), "Number of paths and texts must match"

        questions = [question_template.format(text) for text in texts]
        answers = [answer_template.format(text) for text in texts]

        processed_data = self.load_images(paths, num_segments=num_frames)

        lm_probs = []
        for data, question, answer in zip(processed_data, questions, answers):
            data = data.to(self.device)
            chat_history = []
            media_type='video' if data.dim() == 6 else 'image'  # 6D for video, 5D for image
            instruction=None
            msg = ''
            user_prompt = question
            media_tensor = data
            if self.model_name == 'internvideo2-chat-8b':
                input_ids, attention_masks, labels = [], [], []

                conversation = ""
                if instruction:
                    conversation += instruction
                conversation += (
                            "[INST]" + " "
                        )

                if media_type == 'image':
                    conversation +=( "<Image>" + IMG_TOKEN + "</Image>")#*ilen
                else:
                    conversation += ("<Video>" + VID_TOKEN + "</Video>")#*ilen


                conversation += (
                            msg.rstrip() + "[/INST]"
                        )

                for q,a in chat_history:
                    conversation += (" [INST] " + q + " [/INST]")
                    conversation += (a + "</s>")

                conversation += (" [INST] " + user_prompt + " [/INST]")
                conversation += ("")


                total_len = 0
                indexs = []
                tokenized = self.build_input_ids_chat(
                    self.tokenizer,
                    conversation,
                    max_length=248,
                    add_special_tokens=True,
                    truncation=False,
                    padding=False,
                    return_tensors='pt'
                )
                if media_type == 'image':
                    response = self.generate_caption_chat(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        image_idx = tokenized['index'].unsqueeze(0),
                        image = media_tensor.unsqueeze(0), 
                       )
                else:
                    response = self.generate_caption_chat(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        video_idx = tokenized['index'].unsqueeze(0),
                        video = media_tensor.unsqueeze(0), 
                       )
            elif self.model_name == 'internvideo2-chat-8b-hd':
                answer_prompt = None
                conversation = ""
                if instruction:
                    conversation += instruction
                conversation += (
                            "[INST]" + " "
                        )

                if media_type == 'image':
                    ilen = media_tensor.shape[0]
                    conversation +=( "<Image>" + IMG_TOKEN + "</Image>")*ilen
                else:
                    ilen = media_tensor.shape[1]
                    conversation += ("<Video>" + VID_TOKEN + "</Video>")*ilen


                conversation += (
                            msg.rstrip() + "[/INST]"
                        )

                for q,a in chat_history:
                    conversation += (" [INST] " + q + " [/INST]")
                    conversation += (a + "</s>")

                conversation += (" [INST] " + user_prompt + " [/INST]")
                if answer_prompt:
                    conversation += (answer_prompt)
                else:
                    conversation += ("")


                total_len = 0
                indexs = []
                # ilen = media_tensor.shape[1]

                # conversation = ""
                # if instruction:
                #     conversation += instruction
                # conversation += (
                #             "[INST]" + " "
                #         )

                # if media_type == 'image':
                #     conversation +=( "<Image>" + IMG_TOKEN + "</Image>")*ilen
                # else:
                #     conversation += ("<Video>" + VID_TOKEN + "</Video>")*ilen


                # conversation += (
                #             msg.rstrip() + "[/INST]"
                #         )

                # for q,a in chat_history:
                #     conversation += (" [INST] " + q + " [/INST]")
                #     conversation += (a + "</s>")

                # conversation += (" [INST] " + user_prompt + " [/INST]")
                # conversation += ("")


                # total_len = 0
                # indexs = []
                tokenized = self.build_input_ids_chat_hd(
                    self.tokenizer,
                    conversation,
                    max_length=1024,
                    add_special_tokens=True,
                    truncation=False,
                    padding=False,
                    return_tensors='pt'
                )
                if media_type == 'image':
                    response = self.generate_caption_chat_hd(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        image_idx = tokenized['index'].unsqueeze(0),
                        image = media_tensor,
                        instruction=[instruction]* ilen if instruction else None,
                 
                    )
                else:
                    response = self.generate_caption_chat_hd(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        video_idx = tokenized['index'].unsqueeze(0),
                        video = media_tensor, 
                        instruction=[instruction]* ilen if instruction else None,
                    )
   
            scores = response.scores[0] 

            probs = torch.nn.functional.softmax(scores, dim=-1)
            yes_token_id = self.tokenizer.encode(answer)[1]
            lm_prob = probs[0, yes_token_id].item()
            lm_probs.append(lm_prob)

        return torch.tensor(lm_probs)

    def generate(self,
                    images: List[str],
                    texts: List[str],
                    num_frames: int=8,
                    resolution: int = None,
                    hd_num: int = None) -> torch.Tensor:
        assert len(images) == len(texts), "Number of paths and texts must match"

        questions = texts
        processed_data = self.load_images(images, num_segments=num_frames)

        gen_outputs = []
        for data, question in zip(processed_data, questions):
            data = data.to(self.device)
            chat_history = []
            media_type='video' if data.dim() == 6 else 'image'  # 6D for video, 5D for image
            instruction=None
            msg = ''
            user_prompt = question
            media_tensor = data
            if self.model_name == 'internvideo2-chat-8b':
                input_ids, attention_masks, labels = [], [], []

                conversation = ""
                if instruction:
                    conversation += instruction
                conversation += (
                            "[INST]" + " "
                        )

                if media_type == 'image':
                    conversation +=( "<Image>" + IMG_TOKEN + "</Image>")#*ilen
                else:
                    conversation += ("<Video>" + VID_TOKEN + "</Video>")#*ilen


                conversation += (
                            msg.rstrip() + "[/INST]"
                        )

                for q,a in chat_history:
                    conversation += (" [INST] " + q + " [/INST]")
                    conversation += (a + "</s>")

                conversation += (" [INST] " + user_prompt + " [/INST]")
                conversation += ("")


                total_len = 0
                indexs = []
                tokenized = self.build_input_ids_chat(
                    self.tokenizer,
                    conversation,
                    max_length=248,
                    add_special_tokens=True,
                    truncation=False,
                    padding=False,
                    return_tensors='pt'
                )
                if media_type == 'image':
                    response = self.generate_caption_chat(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        image_idx = tokenized['index'].unsqueeze(0),
                        image = media_tensor.unsqueeze(0), 
                        max_new_tokens=256
                    )
                else:
                    response = self.generate_caption_chat(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        video_idx = tokenized['index'].unsqueeze(0),
                        video = media_tensor.unsqueeze(0), 
                        max_new_tokens=256
                    )
            elif self.model_name == 'internvideo2-chat-8b-hd':
                answer_prompt = None
                conversation = ""
                if instruction:
                    conversation += instruction
                conversation += (
                            "[INST]" + " "
                        )

                if media_type == 'image':
                    ilen = media_tensor.shape[0]
                    conversation +=( "<Image>" + IMG_TOKEN + "</Image>")*ilen
                else:
                    ilen = media_tensor.shape[1]
                    conversation += ("<Video>" + VID_TOKEN + "</Video>")*ilen


                conversation += (
                            msg.rstrip() + "[/INST]"
                        )

                for q,a in chat_history:
                    conversation += (" [INST] " + q + " [/INST]")
                    conversation += (a + "</s>")

                conversation += (" [INST] " + user_prompt + " [/INST]")
                if answer_prompt:
                    conversation += (answer_prompt)
                else:
                    conversation += ("")


                total_len = 0
                indexs = []
                # ilen = media_tensor.shape[1]

                # conversation = ""
                # if instruction:
                #     conversation += instruction
                # conversation += (
                #             "[INST]" + " "
                #         )

                # if media_type == 'image':
                #     conversation +=( "<Image>" + IMG_TOKEN + "</Image>")*ilen
                # else:
                #     conversation += ("<Video>" + VID_TOKEN + "</Video>")*ilen


                # conversation += (
                #             msg.rstrip() + "[/INST]"
                #         )

                # for q,a in chat_history:
                #     conversation += (" [INST] " + q + " [/INST]")
                #     conversation += (a + "</s>")

                # conversation += (" [INST] " + user_prompt + " [/INST]")
                # conversation += ("")


                # total_len = 0
                # indexs = []
                tokenized = self.build_input_ids_chat_hd(
                    self.tokenizer,
                    conversation,
                    max_length=1024,
                    add_special_tokens=True,
                    truncation=False,
                    padding=False,
                    return_tensors='pt'
                )
                if media_type == 'image':
                    response = self.generate_caption_chat_hd(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        image_idx = tokenized['index'].unsqueeze(0),
                        image = media_tensor,
                        instruction=[instruction]* ilen if instruction else None,
                        max_new_tokens=256
                
                    )
                else:
                    response = self.generate_caption_chat_hd(
                        tokenized['input_ids'].unsqueeze(0).to(self.device), 
                        tokenized['attention_mask'].unsqueeze(0).to(self.device), 
                        video_idx = tokenized['index'].unsqueeze(0),
                        video = media_tensor, 
                        instruction=[instruction]* ilen if instruction else None,
                        max_new_tokens=256
                    )
       
    
            tokens = response.sequences[0] 
            text = self.tokenizer.decode(tokens, skip_special_tokens=True).strip()

            gen_outputs.append(text)

        return gen_outputs
        
    # **
    # **
    # Helper Functions:
    # **
    # **

    def load_video_chat(self, video_path, num_segments=8, return_msg=False, resolution=224, hd_num=4, padding=False):
        vr = VideoReader(video_path, ctx=cpu(0), num_threads=1)
        num_frames = len(vr)
        frame_indices = self.get_index(num_frames, num_segments)

        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)

        transform = transforms.Compose([
            transforms.Lambda(lambda x: x.float().div(255.0)),
            transforms.Resize(224, interpolation=transforms.InterpolationMode.BICUBIC),
            transforms.CenterCrop(224),
            transforms.Normalize(mean, std)
        ])

        frames = vr.get_batch(frame_indices)
        frames = frames.permute(0, 3, 1, 2)
        frames = transform(frames)

        T_, C, H, W = frames.shape
            
        if return_msg:
            fps = float(vr.get_avg_fps())
            sec = ", ".join([str(round(f / fps, 1)) for f in frame_indices])
            # " " should be added in the start and end
            msg = f"The video contains {len(frame_indices)} frames sampled at {sec} seconds."
            return frames, msg
        else:
            return frames
        
    def load_video_chat_hd(self, video_path, num_segments=8, return_msg=False, resolution=224, hd_num=4, padding=False):
        vr = VideoReader(video_path, ctx=cpu(0), num_threads=1)
        num_frames = len(vr)
        frame_indices = self.get_index(num_frames, num_segments)

        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)

        transform = transforms.Compose([
            transforms.Lambda(lambda x: x.float().div(255.0)),
            transforms.Normalize(mean, std)
        ])

        frames = vr.get_batch(frame_indices)
        frames = frames.permute(0, 3, 1, 2)

        if padding:
            frames = self.HD_transform_padding(frames.float(), image_size=resolution, hd_num=hd_num)
        else:
            frames = self.HD_transform_no_padding(frames.float(), image_size=resolution, hd_num=hd_num)

        frames = transform(frames)
        T_, C, H, W = frames.shape

        sub_img = frames.reshape(
            1, T_, 3, H//resolution, resolution, W//resolution, resolution
        ).permute(0, 3, 5, 1, 2, 4, 6).reshape(-1, T_, 3, resolution, resolution).contiguous()

        glb_img = F.interpolate(
            frames.float(), size=(resolution, resolution), mode='bicubic', align_corners=False
        ).to(sub_img.dtype).unsqueeze(0)

        frames = torch.cat([sub_img, glb_img]).unsqueeze(0)

        if return_msg:
            fps = float(vr.get_avg_fps())
            sec = ", ".join([str(round(f / fps, 1)) for f in frame_indices])
            # " " should be added in the start and end
            msg = f"The video contains {len(frame_indices)} frames sampled at {sec} seconds."
            return frames, msg
        else:
            return frames

    # *
    # HD Video Loading Helper Functions
    # * 

    def HD_transform_padding(self, frames, image_size=224, hd_num=6):
        def _padding_224(frames):
            _, _, H, W = frames.shape
            tar = int(np.ceil(H / 224) * 224)
            top_padding = (tar - H) // 2
            bottom_padding = tar - H - top_padding
            left_padding = 0
            right_padding = 0

            padded_frames = F.pad(
                frames,
                pad=[left_padding, right_padding, top_padding, bottom_padding],
                mode='constant', value=255
            )
            return padded_frames

        _, _, H, W = frames.shape
        trans = False
        if W < H:
            frames = frames.flip(-2, -1)
            trans = True
            width, height = H, W
        else:
            width, height = W, H

        ratio = width / height
        scale = 1
        while scale * np.ceil(scale / ratio) <= hd_num:
            scale += 1
        scale -= 1
        new_w = int(scale * image_size)
        new_h = int(new_w / ratio)

        resized_frames = F.interpolate(
            frames, size=(new_h, new_w),
            mode='bicubic',
            align_corners=False
        )
        padded_frames = _padding_224(resized_frames)

        if trans:
            padded_frames = padded_frames.flip(-2, -1)

        return padded_frames

    def find_closest_aspect_ratio(self, aspect_ratio, target_ratios, width, height, image_size):
            best_ratio_diff = float('inf')
            best_ratio = (1, 1)
            area = width * height
            for ratio in target_ratios:
                target_aspect_ratio = ratio[0] / ratio[1]
                ratio_diff = abs(aspect_ratio - target_aspect_ratio)
                if ratio_diff < best_ratio_diff:
                    best_ratio_diff = ratio_diff
                    best_ratio = ratio
                elif ratio_diff == best_ratio_diff:
                    if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                        best_ratio = ratio
            return best_ratio


    def HD_transform_no_padding(self, frames, image_size=224, hd_num=6, fix_ratio=(2,1)):
        min_num = 1
        max_num = hd_num
        _, _, orig_height, orig_width = frames.shape
        aspect_ratio = orig_width / orig_height

        # calculate the existing video aspect ratio
        target_ratios = set(
            (i, j) for n in range(min_num, max_num + 1) for i in range(1, n + 1) for j in range(1, n + 1) if
            i * j <= max_num and i * j >= min_num)
        target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])

        # find the closest aspect ratio to the target
        if fix_ratio:
            target_aspect_ratio = fix_ratio
        else:
            target_aspect_ratio = self.find_closest_aspect_ratio(
                aspect_ratio, target_ratios, orig_width, orig_height, image_size)

        # calculate the target width and height
        target_width = image_size * target_aspect_ratio[0]
        target_height = image_size * target_aspect_ratio[1]
        blocks = target_aspect_ratio[0] * target_aspect_ratio[1]

        # resize the frames
        resized_frame = F.interpolate(
            frames, size=(target_height, target_width),
            mode='bicubic', align_corners=False
        )
        return resized_frame
    def process_image_chat(self, image_path, resolution=224):
        mean = (0.48145466, 0.4578275, 0.40821073)
        std = (0.26862954, 0.26130258, 0.27577711)
        d_hid = self.model.vision_encoder.encoder.img_pos_embed.shape[-1]  # Get embedding dimension

        if image_path.lower().endswith('.npy'):
            img = np.load(image_path)
            img = torch.from_numpy(img)
        else:
            img = Image.open(image_path).convert('RGB')
            transform = transforms.Compose([
                transforms.Resize((resolution, resolution), 
                            interpolation=InterpolationMode.BICUBIC),
                transforms.ToTensor(),
                transforms.Normalize(mean, std)
            ])
            img = transform(img)

        img = img.unsqueeze(0).unsqueeze(0)  # Add batch and frame dimensions
        
        # For single image, we can use ckpt_num_frame=-1 to skip interpolation
        n_position = (resolution//16)**2
        new_pos_emb = self.get_sinusoid_encoding_table(
            n_position=n_position,
            d_hid=d_hid,
            ckpt_num_frame=-1,  # Skip frame interpolation for single image
            cur_frame=1  # Single frame
        )
        self.model.vision_encoder.encoder.img_pos_embed = new_pos_emb

        return img.to(self.device)

    def process_image_chat_hd(self, image_path, resolution=224, hd_num=12, padding=False):
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)

        if image_path.lower().endswith('.npy'):
            img = np.load(image_path)
            img = torch.from_numpy(img)
        else:
            img = Image.open(image_path).convert('RGB')
            img = PILToTensor()(img)
        
        # Match authors' implementation
        img = img.float().div(255.0).unsqueeze(0)  # Adding batch dimension first

        if padding:
            img = self.HD_transform_padding(img, image_size=resolution, hd_num=hd_num)
        else:
            img = self.HD_transform_no_padding(img, image_size=resolution, hd_num=hd_num)

        transform = transforms.Compose([
            transforms.Normalize(mean, std)
        ])
        
        img = transform(img).unsqueeze(0)  # Final unsqueeze for model input
        return img.to(self.device)
    
    def get_sinusoid_encoding_table(self, n_position, d_hid, ckpt_num_frame=-1, cur_frame=12): 
        ''' Sinusoid position encoding table ''' 
        # TODO: make it with torch instead of numpy 
        def get_position_angle_vec(position): 
            return [position / np.power(10000, 2 * (hid_j // 2) / d_hid) for hid_j in range(d_hid)] 
        
        if ckpt_num_frame != -1 and ckpt_num_frame != cur_frame:
            logger.info(f"Interpolate position embedding")
            logger.info(f"Testing frame: {cur_frame}")
            logger.info(f"Checkpoint frame: {ckpt_num_frame}")

            T = ckpt_num_frame # checkpoint frame
            new_T = cur_frame # testing frame
            n_position = n_position // new_T * T # generate checkpoint position embedding
            sinusoid_table = np.array([get_position_angle_vec(pos_i) for pos_i in range(n_position)]) 
            sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2]) # dim 2i 
            sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2]) # dim 2i+1 
            sinusoid_table = torch.tensor(sinusoid_table, dtype=torch.float, requires_grad=False).unsqueeze(0)
            # interpolate
            P = int((n_position // T) ** 0.5)
            C = d_hid
            sinusoid_table = sinusoid_table.reshape(-1, T, P, P, C)
            sinusoid_table = sinusoid_table.permute(0, 2, 3, 4, 1).reshape(-1, C, T)  # BHW, C, T
            sinusoid_table = torch.nn.functional.interpolate(sinusoid_table, size=new_T, mode='linear')
            sinusoid_table = sinusoid_table.reshape(1, P, P, C, new_T).permute(0, 4, 1, 2, 3) # B, T, H, W, C
            sinusoid_table = sinusoid_table.flatten(1, 3)
            return sinusoid_table
        else:
            sinusoid_table = np.array([get_position_angle_vec(pos_i) for pos_i in range(n_position)]) 
            sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2]) # dim 2i 
            sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2]) # dim 2i+1 
            return torch.tensor(sinusoid_table, dtype=torch.float, requires_grad=False).unsqueeze(0) 
    def get_index(self, num_frames, num_segments):
        seg_size = float(num_frames - 1) / num_segments
        start = int(seg_size / 2)
        offsets = np.array([
            start + int(np.round(seg_size * idx)) for idx in range(num_segments)
        ])
        return offsets


    # *
    # Chat Helper Functions: Note, there is probably a more efficient way to do this, but this is the most foolproof way.
    # *

    def generate_caption_chat_hd(
        self,
        input_ids,
        attention_mask,
        image_idx = None,
        video_idx = None,
        image: Optional[torch.Tensor] = None,
        video: Optional[torch.Tensor] = None,
        num_beams=1,
        max_new_tokens=1,
        do_sample=True,
        top_p=0.9,
        top_k=None,
        temperature=1.0,
        length_penalty=1,
        repetition_penalty=1.0,
        instruction=None
    ):
       
        text_embeds = self.model.pad_text_embeds(input_ids=input_ids, image=image, video=video, image_idx=image_idx, video_idx=video_idx,instruction=instruction)
        
        # outputs = self.lm.generate(
        #     inputs_embeds=text_embeds,
        #     attention_mask=attention_mask,
        #     num_beams=num_beams,
        #     max_new_tokens=max_new_tokens,
        #     do_sample=do_sample,
        #     min_length=1,
        #     top_p=top_p,
        #     top_k=top_k,
        #     temperature=temperature,
        #     length_penalty=length_penalty,
        #     repetition_penalty=repetition_penalty,
        # )
        # Ignoring InternVideo2's default parameters
        outputs = self.model.lm.generate(
            inputs_embeds=text_embeds,
            attention_mask=attention_mask,
            num_beams=num_beams,
            max_new_tokens=max_new_tokens,
            do_sample=False,
            output_scores=True, 
            return_dict_in_generate=True
        )

        return outputs
    def build_input_ids_chat_hd(
            self, 
            tokenizer, 
            conversation,
            max_length,
            add_special_tokens,
            truncation,
            image = None, 
            video = None, 
            padding = "longest", 
            return_tensors = "pt",
            image_placeholder: str = DEFAULT_IMG_PLACEHOLDER,
            video_placeholder: str = DEFAULT_VID_PLACEHOLDER,
    ):
        input_ids = []
        indexs = []
        attention_mask = []
        start, total_len = 0, 0
        while True:
            index1 = conversation.find(image_placeholder, start)
            index2 = conversation.find(video_placeholder, start)
            if index1 == -1 and index2 == -1:
                index = -1
            elif index1 == -1:
                index = index2
            elif index2 == -1:
                index = index1
            else:
                index = min(index1, index2)
                assert index != -1
            if index == -1:
                inputs = tokenizer(conversation[start:], max_length=max_length-total_len, truncation=truncation, padding=padding, return_tensors=return_tensors)
            else:
                inputs = tokenizer(conversation[start:index], max_length=max_length,  truncation=truncation, padding='longest', return_tensors=return_tensors)
            
            input_ids += inputs.input_ids
            attention_mask += inputs.attention_mask
            total_len += inputs.input_ids[0].shape[0]
            indexs += torch.zeros_like(inputs.input_ids)
            
            if index != -1:
                input_ids += [torch.zeros(96).long()]
                attention_mask += [torch.ones(96).long()]
                indexs += [torch.ones(96)]
            
            if index == -1:
                return {
                    'input_ids': torch.cat(input_ids),
                    'attention_mask': torch.cat(attention_mask),
                    'index': torch.cat(indexs).to(torch.bool),
                }
            start = index + len(DEFAULT_IMG_PLACEHOLDER)

    def generate_caption_chat(
        self,
        input_ids,
        attention_mask,
        image_idx = None,
        video_idx = None,
        image: Optional[torch.Tensor] = None,
        video: Optional[torch.Tensor] = None,
        num_beams=1,
        max_new_tokens=1,
        do_sample=True,
        top_p=0.9,
        top_k=None,
        temperature=1.0,
        length_penalty=1,
        repetition_penalty=1.0,
    ):
        
        text_embeds = self.model.pad_text_embeds(input_ids=input_ids, image=image, video=video, image_idx=image_idx, video_idx=video_idx)
        # outputs = self.lm.generate(
        #     inputs_embeds=text_embeds,
        #     attention_mask=attention_mask,
        #     num_beams=num_beams,
        #     max_new_tokens=max_new_tokens,
        #     do_sample=do_sample,
        #     min_length=1,
        #     top_p=top_p,
        #     top_k=top_k,
        #     temperature=temperature,
        #     length_penalty=length_penalty,
        #     repetition_penalty=repetition_penalty,
        # )
        # Ignoring InternVideo2's default parameters
        outputs = self.model.lm.generate(
            inputs_embeds=text_embeds,
            attention_mask=attention_mask,
            num_beams=num_beams,
            max_new_tokens=max_new_tokens,
            do_sample=False,
            output_scores=True, 
            return_dict_in_generate=True
        )
        return outputs
    
    def build_input_ids_chat(
            self, 
            tokenizer, 
            conversation,
            max_length,
            add_special_tokens,
            truncation,
            image = None, 
            video = None, 
            padding = "longest", 
            return_tensors = "pt",
            image_placeholder: str = DEFAULT_IMG_PLACEHOLDER,
            video_placeholder: str = DEFAULT_VID_PLACEHOLDER,
    ):
        input_ids = []
        indexs = []
        attention_mask = []
        start, total_len = 0, 0
        while True:
            index1 = conversation.find(image_placeholder, start)
            index2 = conversation.find(video_placeholder, start)
            if index1 == -1 and index2 == -1:
                index = -1
            elif index1 == -1:
                index = index2
            elif index2 == -1:
                index = index1
            else:
                index = min(index1, index2)
                assert index != -1
            if index == -1:
                inputs = tokenizer(conversation[start:], max_length=max_length-total_len, truncation=truncation, padding=padding, return_tensors=return_tensors)
            else:
                inputs = tokenizer(conversation[start:index], max_length=max_length,  truncation=truncation, padding='longest', return_tensors=return_tensors)
            
            input_ids += inputs.input_ids
            attention_mask += inputs.attention_mask
            total_len += inputs.input_ids[0].shape[0]
            indexs += torch.zeros_like(inputs.input_ids)
            
            if index != -1:
                input_ids += [torch.zeros(96).long()]
                attention_mask += [torch.ones(96).long()]
                indexs += [torch.ones(96)]
            
            if index == -1:
                return {
                    'input_ids': torch.cat(input_ids),
                    'attention_mask': torch.cat(attention_mask),
                    'index': torch.cat(indexs).to(torch.bool),
                }
            start = index + len(DEFAULT_IMG_PLACEHOLDER)


"""Lip Sync Error — LSE-D / LSE-C (SyncNet / Wav2Lip, 2020).

Measures audio-visual lip synchronisation quality:
  LSE-D (Lip Sync Error - Distance): lower = better sync
  LSE-C (Lip Sync Error - Confidence): higher = better sync

Uses SyncNet-based scoring or falls back to audio-video temporal
correlation heuristic.

pip install syncnet (or wav2lip)
"""

import logging
import subprocess
import tempfile
import cv2
import numpy as np
from pathlib import Path
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class LipSyncModule(PipelineModule):
    name = "lip_sync"
    description = "LSE-D/LSE-C lip sync error (SyncNet/Wav2Lip, 2020)"
    default_config = {
        "subsample": 16,
        "sample_rate": 16000,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._ml_available = False
        self._backend = None
        self.subsample = self.config.get("subsample", 16)
        self.sample_rate = self.config.get("sample_rate", 16000)

    def setup(self) -> None:
        # Tier 1: syncnet package
        try:
            import syncnet
            self._model = syncnet
            self._ml_available = True
            self._backend = "syncnet"
            logger.info("Lip Sync module initialised (syncnet package)")
            return
        except ImportError:
            pass
        except Exception as e:
            logger.debug(f"syncnet init failed: {e}")

        # Tier 2: heuristic (audio-video temporal correlation)
        logger.info("Lip Sync module initialised (heuristic fallback)")

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample

        try:
            if self._ml_available and self._backend == "syncnet":
                scores = self._score_syncnet(sample.path)
            else:
                scores = self._score_heuristic(sample.path)

            if scores is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.lse_d = scores.get("lse_d")
            sample.quality_metrics.lse_c = scores.get("lse_c")
            logger.debug(
                f"Lip Sync for {sample.path.name}: "
                f"LSE-D={scores.get('lse_d', 0.0):.4f} LSE-C={scores.get('lse_c', 0.0):.4f}"
            )
        except Exception as e:
            logger.error(f"Lip Sync failed: {e}")
        return sample

    def _score_syncnet(self, path: Path) -> Optional[dict]:
        try:
            result = self._model.evaluate(str(path))
            return {
                "lse_d": float(result.get("lse_d", result.get("distance", 0))),
                "lse_c": float(result.get("lse_c", result.get("confidence", 0))),
            }
        except Exception as e:
            logger.debug(f"SyncNet scoring failed: {e}")
            return self._score_heuristic(path)

    def _score_heuristic(self, path: Path) -> Optional[dict]:
        """Heuristic: correlate mouth-region visual activity with audio energy.

        LSE-D proxy: inverse correlation between mouth motion and audio energy.
        LSE-C proxy: peak cross-correlation value.
        """
        # Extract visual mouth-region activity
        mouth_energy = self._extract_mouth_energy(path)
        if mouth_energy is None or len(mouth_energy) < 5:
            return None

        # Extract audio energy
        audio_energy = self._extract_audio_energy(path, len(mouth_energy))
        if audio_energy is None or len(audio_energy) < 5:
            return None

        # Align lengths
        min_len = min(len(mouth_energy), len(audio_energy))
        v_sig = mouth_energy[:min_len].astype(np.float64)
        a_sig = audio_energy[:min_len].astype(np.float64)

        # Normalise
        v_sig -= v_sig.mean()
        a_sig -= a_sig.mean()
        v_std = v_sig.std()
        a_std = a_sig.std()

        if v_std < 1e-8 or a_std < 1e-8:
            return {"lse_d": 10.0, "lse_c": 0.0}

        v_sig /= v_std
        a_sig /= a_std

        # Cross-correlation
        corr = np.correlate(a_sig, v_sig, mode="full")
        corr /= min_len

        peak_corr = float(np.max(corr))
        best_lag = int(np.argmax(corr)) - (min_len - 1)

        # LSE-D: distance proxy (lower = better sync)
        # Map: high correlation at small lag → low distance
        lse_d = max(0.0, 1.0 - peak_corr) * 10.0 + abs(best_lag) * 0.5

        # LSE-C: confidence proxy (higher = better sync)
        lse_c = float(np.clip(peak_corr, 0.0, 1.0)) * 10.0

        return {"lse_d": lse_d, "lse_c": lse_c}

    def _extract_mouth_energy(self, path: Path) -> Optional[np.ndarray]:
        """Extract per-frame mouth region motion energy."""
        cap = cv2.VideoCapture(str(path))
        try:
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total < 2:
                return None

            n_sample = min(self.subsample, total)
            indices = np.linspace(0, total - 1, n_sample, dtype=int)

            face_cascade = cv2.CascadeClassifier(
                cv2.data.haarcascades + "haarcascade_frontalface_default.xml"
            )

            prev_mouth = None
            energies = []

            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if not ret:
                    continue

                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                faces = face_cascade.detectMultiScale(gray, 1.3, 5)

                if len(faces) > 0:
                    x, y, w, h = max(faces, key=lambda f: f[2] * f[3])
                    # Lower half of face = mouth region
                    mouth_y = y + int(h * 0.6)
                    mouth_roi = gray[mouth_y:y + h, x:x + w]
                    if mouth_roi.size > 0:
                        mouth_resized = cv2.resize(mouth_roi, (32, 16)).astype(np.float64)
                    else:
                        mouth_resized = np.zeros((16, 32), dtype=np.float64)
                else:
                    # Fallback: use bottom-center region
                    fh, fw = gray.shape
                    mouth_resized = cv2.resize(
                        gray[int(fh * 0.6):fh, int(fw * 0.25):int(fw * 0.75)],
                        (32, 16)
                    ).astype(np.float64)

                if prev_mouth is not None:
                    energy = float(np.mean(np.abs(mouth_resized - prev_mouth)))
                    energies.append(energy)
                else:
                    energies.append(0.0)

                prev_mouth = mouth_resized

            return np.array(energies) if energies else None
        finally:
            cap.release()

    def _extract_audio_energy(self, path: Path, n_frames: int) -> Optional[np.ndarray]:
        """Extract per-frame audio energy."""
        try:
            tmp = tempfile.NamedTemporaryFile(suffix=".wav", delete=False)
            tmp.close()
            cmd = [
                "ffmpeg", "-y", "-i", str(path),
                "-vn", "-ac", "1", "-ar", str(self.sample_rate),
                "-sample_fmt", "s16", tmp.name,
            ]
            result = subprocess.run(cmd, capture_output=True, timeout=30)
            if result.returncode != 0:
                Path(tmp.name).unlink(missing_ok=True)
                return None

            import soundfile as sf
            audio, sr = sf.read(tmp.name, dtype="float32")
            Path(tmp.name).unlink(missing_ok=True)

            if audio.ndim > 1:
                audio = audio.mean(axis=1)

            # Compute per-frame energy
            samples_per_frame = len(audio) // n_frames
            if samples_per_frame < 1:
                return None

            energies = []
            for i in range(n_frames):
                start = i * samples_per_frame
                end = start + samples_per_frame
                if end > len(audio):
                    break
                chunk = audio[start:end]
                rms = float(np.sqrt(np.mean(chunk ** 2)))
                energies.append(rms)

            return np.array(energies)
        except Exception as e:
            logger.debug(f"Audio energy extraction failed: {e}")
            return None



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
Backbone modules.
"""
from collections import OrderedDict
from functools import partial
from typing import Dict, List

import torch
import torch.nn.functional as F
import torchvision
from torch import nn
from torch.cuda.amp import autocast
from torchvision.models._utils import IntermediateLayerGetter
from util.misc import is_main_process, NestedTensor

from .position_encoding import build_position_encoding
from .swin import get_swinl
from .pev1 import get_pev1_and_fpn_backbone


class FrozenBatchNorm2d(torch.nn.Module):
    """
    BatchNorm2d where the batch statistics and the affine parameters are fixed.

    Copy-paste from torchvision.misc.ops with added eps before rqsrt,
    without which any other models than torchvision.models.resnet[18,34,50,101]
    produce nans.
    """

    def __init__(self, n, eps=1e-5):
        super(FrozenBatchNorm2d, self).__init__()
        self.register_buffer("weight", torch.ones(n))
        self.register_buffer("bias", torch.zeros(n))
        self.register_buffer("running_mean", torch.zeros(n))
        self.register_buffer("running_var", torch.ones(n))
        self.eps = eps

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        num_batches_tracked_key = prefix + "num_batches_tracked"
        if num_batches_tracked_key in state_dict:
            del state_dict[num_batches_tracked_key]

        super(FrozenBatchNorm2d, self)._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )

    def forward(self, x):
        # move reshapes to the beginning
        # to make it fuser-friendly
        w = self.weight.reshape(1, -1, 1, 1)
        b = self.bias.reshape(1, -1, 1, 1)
        rv = self.running_var.reshape(1, -1, 1, 1)
        rm = self.running_mean.reshape(1, -1, 1, 1)
        eps = self.eps
        scale = w * (rv + eps).rsqrt()
        bias = b - rm * scale
        return x * scale + bias


class BackboneBase(nn.Module):

    def __init__(
        self, backbone: nn.Module, train_backbone: bool, return_interm_layers: bool
    ):
        super().__init__()
        for name, parameter in backbone.named_parameters():
            if (
                not train_backbone
                or "layer2" not in name
                and "layer3" not in name
                and "layer4" not in name
            ):
                parameter.requires_grad_(False)
        if return_interm_layers:
            # return_layers = {"layer1": "0", "layer2": "1", "layer3": "2", "layer4": "3"}
            return_layers = {"layer2": "0", "layer3": "1", "layer4": "2"}
            self.strides = [8, 16, 32]
            self.num_channels = [512, 1024, 2048]
        else:
            return_layers = {"layer4": "0"}
            self.strides = [32]
            self.num_channels = [2048]
        self.body = IntermediateLayerGetter(backbone, return_layers=return_layers)

    def forward(self, tensor_list: NestedTensor):
        xs = self.body(tensor_list.tensors)
        out: Dict[str, NestedTensor] = {}
        for name, x in xs.items():
            m = tensor_list.mask
            assert m is not None
            mask = F.interpolate(m[None].float(), size=x.shape[-2:]).to(torch.bool)[0]
            out[name] = NestedTensor(x, mask)
        return out


class Backbone(BackboneBase):
    """ResNet backbone with frozen BatchNorm."""

    def __init__(
        self,
        name: str,
        train_backbone: bool,
        return_interm_layers: bool,
        dilation: bool,
    ):
        norm_layer = FrozenBatchNorm2d
        backbone = getattr(torchvision.models, name)(
            replace_stride_with_dilation=[False, False, dilation],
            pretrained=is_main_process(),
            norm_layer=norm_layer,
        )
        assert name not in ("resnet18", "resnet34"), "number of channels are hard coded"
        super().__init__(backbone, train_backbone, return_interm_layers)
        if dilation:
            self.strides[-1] = self.strides[-1] // 2


class SwinBackbone(nn.Module):
    def __init__(self):
        # we skip R50 FrozenBatchNorm2d, dilation, train l{2,3,4} only
        super().__init__()
        self.body = get_swinl()
        self.features = ["res3", "res4", "res5"]
        self.strides = [8, 16, 32]
        self.num_channels = [384, 768, 1536]

    def forward(self, tensor_list: NestedTensor):
        xs = self.body(tensor_list.tensors)
        m = tensor_list.mask[None]
        assert m is not None
        out: Dict[str, NestedTensor] = {}
        for name in self.features:
            mask = F.interpolate(m.float(), size=xs[name].shape[-2:]).to(torch.bool)[0]
            out[name] = NestedTensor(xs[name], mask)
        return out


class PEv1Backbone(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.body = get_pev1_and_fpn_backbone(args)
        self.features = self.body._out_features

        self.bf16 = args.bf16
        self.fp16 = args.fp16

        _out_feature_strides = self.body._out_feature_strides
        _out_feature_channels = self.body._out_feature_channels
        self.strides = [_out_feature_strides[f] for f in _out_feature_strides.keys()]
        self.num_channels = [
            _out_feature_channels[f] for f in _out_feature_channels.keys()
        ]

    def forward(self, tensor_list: NestedTensor):
        # xs = self.body(tensor_list.tensors)
        # backbone
        if self.bf16:
            with autocast(dtype=torch.bfloat16):
                xs = self.body(tensor_list.tensors.to(torch.bfloat16))
            xs = {k: v.float() for k, v in xs.items()}
        elif self.fp16:
            with autocast(dtype=torch.float16):
                xs = self.body(tensor_list.tensors.half())
            xs = {k: v.float() for k, v in xs.items()}
        else:
            xs = self.body(tensor_list.tensors)

        m = tensor_list.mask[None]
        assert m is not None
        out: Dict[str, NestedTensor] = {}

        for name in self.features:
            mask = F.interpolate(m.float(), size=xs[name].shape[-2:]).to(torch.bool)[0]
            out[name] = NestedTensor(xs[name], mask)
        return out


class Joiner(nn.Sequential):
    def __init__(self, backbone, position_embedding):
        super().__init__(backbone, position_embedding)
        self.strides = backbone.strides
        self.num_channels = backbone.num_channels

    def forward(self, tensor_list: NestedTensor):
        xs = self[0](tensor_list)
        out: List[NestedTensor] = []
        pos = []
        for name, x in sorted(xs.items()):
            out.append(x)

        # position encoding
        for x in out:
            pos.append(self[1](x).to(x.tensors.dtype))

        return out, pos


def build_backbone(args):
    position_embedding = build_position_encoding(args)
    train_backbone = args.lr_backbone > 0
    return_interm_layers = args.masks or (args.num_feature_levels > 1)
    if "swin" in args.backbone:
        backbone = SwinBackbone()
    elif "pev1" in args.backbone:
        backbone = PEv1Backbone(args)
    else:
        backbone = Backbone(
            args.backbone, train_backbone, return_interm_layers, args.dilation
        )
    model = Joiner(backbone, position_embedding)
    return model



"""pVMAF — Predictive VMAF (35x faster).

2024 — lightweight VMAF approximation using bitstream and pixel-level
features to predict VMAF scores without full reference decoding.
Achieves ~35x speedup with high correlation to standard VMAF.

pvmaf_score — 0-100 scale (higher = better)
"""

import logging
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import ReferenceBasedModule

logger = logging.getLogger(__name__)


class PVMAFModule(ReferenceBasedModule):
    name = "pvmaf"
    description = "Predictive VMAF ~35x faster via bitstream+pixel features (2024, 0-100)"
    metric_field = "pvmaf_score"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try native pVMAF
        try:
            import pvmaf
            self._model = pvmaf
            self._backend = "native"
            logger.info("pVMAF (native) initialised")
            return
        except ImportError:
            pass

        # Tier 2: Heuristic fallback
        self._backend = "heuristic"
        logger.info("pVMAF (heuristic) initialised — install pvmaf for full model")

    def compute_reference_score(
        self, sample_path: Path, reference_path: Path
    ) -> Optional[float]:
        if self._backend == "native":
            score = float(self._model.predict(str(sample_path), str(reference_path)))
            return float(np.clip(score, 0.0, 100.0))
        return self._compute_heuristic(sample_path, reference_path)

    def _read_frames(self, path: Path) -> list:
        """Read frames from video or image."""
        frames = []
        is_video = path.suffix.lower() in {
            ".mp4", ".avi", ".mov", ".mkv", ".webm", ".flv", ".wmv",
        }

        if is_video:
            cap = cv2.VideoCapture(str(path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return frames
                indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        frames.append(frame)
            finally:
                cap.release()
        else:
            img = cv2.imread(str(path))
            if img is not None:
                frames.append(img)

        return frames

    def _compute_heuristic(
        self, sample_path: Path, reference_path: Path
    ) -> Optional[float]:
        """Heuristic: bitstream+pixel features → VMAF estimate (0-100)."""
        dist_frames = self._read_frames(sample_path)
        ref_frames = self._read_frames(reference_path)

        if not dist_frames or not ref_frames:
            return None

        n_frames = min(len(dist_frames), len(ref_frames))
        dist_frames = dist_frames[:n_frames]
        ref_frames = ref_frames[:n_frames]

        frame_scores = []
        for i in range(n_frames):
            dist_gray = cv2.cvtColor(dist_frames[i], cv2.COLOR_BGR2GRAY).astype(np.float64)
            ref_gray = cv2.cvtColor(ref_frames[i], cv2.COLOR_BGR2GRAY).astype(np.float64)

            if dist_gray.shape != ref_gray.shape:
                dist_gray = cv2.resize(dist_gray, (ref_gray.shape[1], ref_gray.shape[0]))

            # VIF (Visual Information Fidelity) component
            eps = 1e-7
            ref_mu = cv2.GaussianBlur(ref_gray, (7, 7), 1.0)
            dist_mu = cv2.GaussianBlur(dist_gray, (7, 7), 1.0)
            ref_var = cv2.GaussianBlur(ref_gray ** 2, (7, 7), 1.0) - ref_mu ** 2
            dist_var = cv2.GaussianBlur(dist_gray ** 2, (7, 7), 1.0) - dist_mu ** 2
            cross_var = cv2.GaussianBlur(ref_gray * dist_gray, (7, 7), 1.0) - ref_mu * dist_mu

            ref_var = np.maximum(ref_var, 0)
            dist_var = np.maximum(dist_var, 0)

            g = cross_var / (ref_var + eps)
            noise_var = dist_var - g * cross_var
            noise_var = np.maximum(noise_var, eps)

            sigma_n = 2.0
            vif_num = np.sum(np.log2(1 + g ** 2 * ref_var / (noise_var + sigma_n ** 2) + eps))
            vif_den = np.sum(np.log2(1 + ref_var / (sigma_n ** 2) + eps))
            vif = float(vif_num / (vif_den + eps))

            # DLM (Detail Loss Metric) component
            ref_lap = cv2.Laplacian(ref_gray, cv2.CV_64F)
            dist_lap = cv2.Laplacian(dist_gray, cv2.CV_64F)
            detail_loss = np.mean(np.abs(ref_lap - dist_lap)) / (np.mean(np.abs(ref_lap)) + eps)
            dlm = 1.0 / (1.0 + detail_loss)

            # Motion component (pixel-level)
            # Simple MSE-based
            mse = np.mean((ref_gray - dist_gray) ** 2)
            psnr_norm = min(10.0 * np.log10(255.0 ** 2 / (mse + eps)) / 50.0, 1.0) if mse > 0 else 1.0

            # VMAF-like combination (learned weights approximation)
            vmaf_raw = 0.40 * vif + 0.35 * dlm + 0.25 * psnr_norm
            frame_scores.append(vmaf_raw)

        raw_score = float(np.mean(frame_scores))
        # Map to 0-100 VMAF scale
        score = raw_score * 100.0
        return float(np.clip(score, 0.0, 100.0))



"""Dynamics Range module.

Measures the extent of variations in video content (DEVIL protocol).
Assesses how much motion and scene variation occurs in the video.
Range: 0-100 (higher = more dynamic content).
"""

import logging

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class DynamicsRangeModule(PipelineModule):
    name = "dynamics_range"
    description = "Measures extent of motion and content variation (DEVIL protocol)"
    default_config = {
        # Pure OpenCV
        "scene_change_threshold": 30.0,  # Threshold for detecting scene changes
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.scene_change_threshold = self.config.get("scene_change_threshold", 30.0)

    def setup(self) -> None:
        pass  # No setup needed

    def process(self, sample: Sample) -> Sample:
        """Process sample to compute dynamics range."""
        if not sample.is_video:
            return sample  # Dynamics range only for videos

        try:
            cap = cv2.VideoCapture(str(sample.path))
            if not cap.isOpened():
                return sample

            max_frames = self.config.get("max_frames", 300)
            total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

            # Single-pass: read all frames once, compute all three components
            diffs = []
            magnitudes = []
            scene_changes = 0
            prev_gray = None
            frame_count = 0

            while True:
                ret, frame = cap.read()
                if not ret:
                    break

                if frame_count >= max_frames:
                    break

                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

                if prev_gray is not None:
                    # Component 1: frame difference (every pair)
                    diff = cv2.absdiff(prev_gray, gray).mean()
                    diffs.append(diff)

                    # Component 3: scene change detection (every pair)
                    if diff > self.scene_change_threshold:
                        scene_changes += 1

                    # Component 2: optical flow (every 5th frame)
                    if frame_count % 5 == 0:
                        try:
                            flow = cv2.calcOpticalFlowFarneback(
                                prev_gray, gray, None, 0.5, 3, 15, 3, 5, 1.2, 0
                            )
                            magnitude, _ = cv2.cartToPolar(flow[..., 0], flow[..., 1])
                            magnitudes.append(magnitude.mean())
                        except Exception:
                            pass

                prev_gray = gray
                frame_count += 1

            cap.release()

            # Derive per-component results
            diff_variance = float(np.var(diffs)) if diffs else 0.0
            motion_range = float(np.max(magnitudes) - np.min(magnitudes)) if magnitudes else 0.0
            scene_change_freq = scene_changes / max(total_frames, 1) if total_frames > 0 else 0

            # Combine into normalized score
            # Normalize each component (empirical ranges)
            diff_var_norm = min(diff_variance / 100.0, 1.0)
            motion_range_norm = min(motion_range / 10.0, 1.0)
            scene_freq_norm = min(scene_change_freq * 100, 1.0)

            dynamics_range = (diff_var_norm + motion_range_norm + scene_freq_norm) / 3.0 * 100.0

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.dynamics_range = dynamics_range

            logger.debug(
                f"Dynamics range for {sample.path.name}: {dynamics_range:.1f} "
                f"(diff_var: {diff_variance:.1f}, motion: {motion_range:.2f}, scenes: {scene_changes})"
            )

        except Exception as e:
            logger.warning(f"Dynamics range processing failed for {sample.path}: {e}")

        return sample



import logging

import torch
import torch.nn.functional as F
from einops import rearrange

from models.utils import tile

from .backbones.vit import build_vit
from .backbones.bert.builder import build_bert_decoder
from .umt import UMT

logger = logging.getLogger(__name__)


class UMT_QA(UMT):
    """docstring for UMT_QA"""

    def __init__(self, config, tokenizer, is_pretrain=False):
        super(UMT_QA, self).__init__(config, tokenizer, is_pretrain)

        # delete extra/unnecessary modules inherited from SingularityRetrievalBase
        extra_attributes = ["vision_proj", "text_proj", "temp", "itm_head"]
        for attr in extra_attributes:
            delattr(self, attr)

        self.text_decoder = self.build_text_decoder()

    def build_vision_encoder(self):
        """build vision encoder
        Returns: (vision_encoder, clip_teacher). Each is a `nn.Module`.

        """
        encoder_name = self.config.model.vision_encoder.name
        logger.info(f"Build vision_encoder: {encoder_name}")
        if "vit" in encoder_name:
            vision_encoder = build_vit(self.config.model, add_pool_norm=False)
        else:
            raise ValueError(f"not implemented: {encoder_name}")
        return vision_encoder

    def build_text_decoder(self):
        encoder_name = self.config.model.text_encoder.name
        logger.info(f"Build text_decoder {encoder_name}")
        if "bert" in encoder_name:
            text_decoder = build_bert_decoder(
                self.config.model, self.config.gradient_checkpointing
            )
        else:
            raise NotImplementedError()
        return text_decoder

    def encode_vision(self, image):
        """encode image / videos as features.

        Args:
            image (torch.Tensor): The input images.
            test (bool): Whether testing.

        Returns: tuple.
            - vision_embeds (torch.Tensor): The output features. Shape: [B,N,C].

        """
        T = image.shape[1]
        use_image = True if T == 1 else False
        image = image.permute(0, 2, 1, 3, 4) # [B,T,C,H,W] -> [B,C,T,H,W]
        # whether save temporal dimension
        keep_temporal = self.config.model.vision_encoder.keep_temporal
        vision_embeds, _ = self.vision_encoder(
            image, None, None, use_image, keep_temporal,
        )
        return vision_embeds

    def forward(self, image, question, answer=None, k=None, weights=None, train=True):
        """
        Args:
        k: number of answers for each question
        weights: weight for each answer
        """
        image_embeds = self.encode_vision(image)
        image_atts = torch.ones(image_embeds.size()[:-1], dtype=torch.long).to(image.device)

        if train:
            answer_targets = answer.input_ids.masked_fill(
                answer.input_ids == self.tokenizer.pad_token_id, -100
            )

            question_output = self.text_encoder(
                question.input_ids,
                attention_mask=question.attention_mask,
                encoder_hidden_states=image_embeds,
                encoder_attention_mask=image_atts,
                return_dict=True,
            )

            question_states = []
            question_atts = []
            for b, n in enumerate(k):
                question_states += [question_output.last_hidden_state[b]] * n
                question_atts += [question.attention_mask[b]] * n
            question_states = torch.stack(question_states, 0)
            question_atts = torch.stack(question_atts, 0)

            answer_output = self.text_decoder(
                answer.input_ids,
                attention_mask=answer.attention_mask,
                encoder_hidden_states=question_states,
                encoder_attention_mask=question_atts,
                labels=answer_targets,
                return_dict=True,
                reduction="none",
            )
            loss = weights * answer_output.loss
            loss = loss.sum() / image.size(0)

            return loss

        # inference
        else:
            question_output = self.text_encoder(
                question.input_ids,
                attention_mask=question.attention_mask,
                encoder_hidden_states=image_embeds,
                encoder_attention_mask=image_atts,
                return_dict=True,
            )
            topk_ids, topk_probs = self.rank_answer(
                question_output.last_hidden_state,
                question.attention_mask,
                answer.input_ids,
                answer.attention_mask,
                k,
            )  # (bsz, 128), (bsz, 128)
            return topk_ids, topk_probs

    def rank_answer(self, question_states, question_atts, answer_ids, answer_atts, k):
        """
        question_states: (bsz, Lq, d)
        answer_ids: answer input id after tokenization, (#answers, La)
        """
        num_ques = question_states.size(0)
        start_ids = answer_ids[0, 0].repeat(num_ques, 1)  # bos token

        start_output = self.text_decoder(
            start_ids,
            encoder_hidden_states=question_states,
            encoder_attention_mask=question_atts,
            return_dict=True,
            reduction="none",
        )
        logits = start_output.logits[:, 0, :]  # first token's logit

        # topk_probs: top-k probability
        # topk_ids: [num_question, k]
        answer_first_token = answer_ids[:, 1]
        prob_first_token = F.softmax(logits, dim=1).index_select(
            dim=1, index=answer_first_token
        )
        topk_probs, topk_ids = prob_first_token.topk(k, dim=1)

        # answer input: [num_question*k, answer_len]
        input_ids = []
        input_atts = []
        for b, topk_id in enumerate(topk_ids):
            input_ids.append(answer_ids.index_select(dim=0, index=topk_id))
            input_atts.append(answer_atts.index_select(dim=0, index=topk_id))
        input_ids = torch.cat(input_ids, dim=0)
        input_atts = torch.cat(input_atts, dim=0)

        targets_ids = input_ids.masked_fill(input_ids == self.tokenizer.pad_token_id, -100)

        # repeat encoder's output for top-k answers
        question_states = tile(question_states, 0, k)
        question_atts = tile(question_atts, 0, k)

        output = self.text_decoder(
            input_ids,
            attention_mask=input_atts,
            encoder_hidden_states=question_states,
            encoder_attention_mask=question_atts,
            labels=targets_ids,
            return_dict=True,
            reduction="none",
        )

        answer_loss = output.loss
        answer_loss = answer_loss.view(input_ids.size(0), -1)

        # topk_prob: first token probability
        topk_probs = topk_probs.view(-1, 1)
        log_probs = torch.cat([topk_probs.log(), -answer_loss], dim=1)

        # re-calculate log probabilities for the answer sequences using chain rule
        log_probs_sum = log_probs.sum(1)
        log_probs_sum = log_probs_sum.view(num_ques, k)

        topk_probs = F.softmax(log_probs_sum, dim=-1)
        # get top-k after re-ranking
        topk_probs, rerank_id = topk_probs.topk(k, dim=1)
        topk_ids = torch.gather(topk_ids, 1, rerank_id)

        return topk_ids, topk_probs



"""ERQA — Edge Restoration Quality Assessment (2022).

Full-reference metric that evaluates edge restoration quality by comparing
edge maps between distorted and reference images/frames.

pip install erqa

erqa_score — 0-1, higher = better edge restoration quality.
"""

import logging
import cv2
import numpy as np
from pathlib import Path
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import ReferenceBasedModule

logger = logging.getLogger(__name__)


class ERQAModule(ReferenceBasedModule):
    name = "erqa"
    description = "ERQA edge restoration quality assessment (FR, 2022)"
    metric_field = "erqa_score"
    default_config = {"subsample": 8}

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._ml_available = False
        self.subsample = self.config.get("subsample", 8)

    def setup(self) -> None:
        # Tier 1: erqa package
        try:
            import erqa as erqa_lib
            self._model = erqa_lib.ERQA()
            self._ml_available = True
            logger.info("ERQA module initialised (erqa package)")
            return
        except ImportError:
            pass
        except Exception as e:
            logger.debug(f"ERQA package init failed: {e}")

        # Tier 2: heuristic (edge detection difference)
        logger.info("ERQA module initialised (heuristic fallback)")

    def compute_reference_score(self, sample_path: Path, reference_path: Path) -> Optional[float]:
        try:
            if str(sample_path).lower().endswith((".mp4", ".avi", ".mov", ".mkv", ".webm")):
                return self._score_video(str(sample_path), str(reference_path))
            else:
                return self._score_image(str(sample_path), str(reference_path))
        except Exception as e:
            logger.warning(f"ERQA failed: {e}")
            return None

    def _score_image(self, sample_p: str, ref_p: str) -> Optional[float]:
        img = cv2.imread(sample_p)
        ref = cv2.imread(ref_p)
        if img is None or ref is None:
            return None

        if self._ml_available and self._model is not None:
            return self._score_erqa_package(img, ref)
        return self._score_heuristic(img, ref)

    def _score_video(self, sample_p: str, ref_p: str) -> Optional[float]:
        cap_s = cv2.VideoCapture(sample_p)
        cap_r = cv2.VideoCapture(ref_p)
        try:
            total = min(
                int(cap_s.get(cv2.CAP_PROP_FRAME_COUNT)),
                int(cap_r.get(cv2.CAP_PROP_FRAME_COUNT)),
            )
            if total <= 0:
                return None
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            scores = []
            for idx in indices:
                cap_s.set(cv2.CAP_PROP_POS_FRAMES, idx)
                cap_r.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret_s, frame_s = cap_s.read()
                ret_r, frame_r = cap_r.read()
                if ret_s and ret_r:
                    if self._ml_available and self._model is not None:
                        s = self._score_erqa_package(frame_s, frame_r)
                    else:
                        s = self._score_heuristic(frame_s, frame_r)
                    if s is not None:
                        scores.append(s)
            return float(np.mean(scores)) if scores else None
        finally:
            cap_s.release()
            cap_r.release()

    def _score_erqa_package(self, img: np.ndarray, ref: np.ndarray) -> Optional[float]:
        try:
            h, w = ref.shape[:2]
            img = cv2.resize(img, (w, h))
            score = self._model(img, ref)
            return float(score)
        except Exception as e:
            logger.debug(f"ERQA package scoring failed: {e}")
            return self._score_heuristic(img, ref)

    def _score_heuristic(self, img: np.ndarray, ref: np.ndarray) -> float:
        """Heuristic: compare Canny edge maps via F1-like agreement."""
        h, w = ref.shape[:2]
        img = cv2.resize(img, (w, h))

        gray_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        gray_ref = cv2.cvtColor(ref, cv2.COLOR_BGR2GRAY)

        edges_img = cv2.Canny(gray_img, 100, 200).astype(np.float64)
        edges_ref = cv2.Canny(gray_ref, 100, 200).astype(np.float64)

        # Dilate edges slightly for tolerance
        kernel = np.ones((3, 3), np.uint8)
        edges_img_d = cv2.dilate(edges_img, kernel, iterations=1)
        edges_ref_d = cv2.dilate(edges_ref, kernel, iterations=1)

        # Precision: how many predicted edges match reference
        pred_count = max(edges_img.sum(), 1.0)
        tp_precision = (edges_img * edges_ref_d).sum()
        precision = tp_precision / pred_count

        # Recall: how many reference edges are recovered
        ref_count = max(edges_ref.sum(), 1.0)
        tp_recall = (edges_ref * edges_img_d).sum()
        recall = tp_recall / ref_count

        if precision + recall < 1e-8:
            return 0.0
        f1 = 2 * precision * recall / (precision + recall)
        return float(np.clip(f1, 0.0, 1.0))



from typing import List
import torch
import copy
import os
from torchvision import transforms

from .vqa_model import VQAScoreModel
from .lavis.models import load_model
from ...constants import HF_CACHE_DIR

default_question_template = 'Question: Does this figure show "{}"? Please answer yes or no.'
default_answer_template = "yes" # instruct-blip uses "yes" instead of "Yes"

InstructBLIP_MODELS = {
    'instructblip-flant5-xxl': {'variant': 'flant5xxl'},
    'instructblip-flant5-xl': {'variant': 'flant5xl'},
}

class InstructBLIPModel(VQAScoreModel):
    """A wrapper for the InstructBLIP (FlanT5-based) models"""
    def __init__(self,
                 model_name='instructblip-flant5-xxl',
                 device='cuda',
                 cache_dir=HF_CACHE_DIR):
        assert model_name in InstructBLIP_MODELS, f"Model name {model_name} not found in InstructBLIP_MODELS"
        os.environ['TORCH_HOME'] = cache_dir
        import timm.models as timm_hub
        video_mode = "concat"
        allows_image = True
        # if cache_dir != timm_hub.get_cache_dir():
        #     print(f"Warning: cache_dir {cache_dir} won't be used because "
        #            "InstructBLIP is cached using timm_hub.get_cache_dir(). "
        #            "You may want to update os.environ['TORCH_HOME'] before importing to specify the cache_dir.")
        #     print(f"Model saved to {timm_hub.get_cache_dir()}")
        super().__init__(model_name=model_name,
                         device=device,
                         cache_dir=cache_dir)

    def load_model(self):
        """Load the model, tokenizer, image transform
        """
        self.variant = InstructBLIP_MODELS[self.model_name]['variant']
        self.model = load_model("blip2_t5_instruct", self.variant, is_eval=True, device=self.device)
        size = 224
        self.image_preprocess = transforms.Compose([
            transforms.Resize((size),interpolation=transforms.functional.InterpolationMode.BICUBIC),
            transforms.CenterCrop((size,size)),
            transforms.ToTensor(),
            transforms.Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711))]
        )  

    def load_images(self,
                    image: List[str]) -> torch.Tensor:
        """Load the image(s), and return a tensor (after preprocessing) put on self.device
        """
        if any(path.startswith(("http://", "https://")) for path in image):
            raise NotImplementedError("Web link image/video inputs are not yet supported for this model. Please use a local path, or otherwise, make a Github issue request if this feature is necessary.")
        image = [self.image_loader(x) for x in image]
        image = [self.image_preprocess(image) for image in image]
        assert all(x.shape == image[0].shape for x in image)
        image = torch.stack(image, dim=0).to(self.device)
        return image

    @torch.no_grad()
    @torch.autocast(device_type='cuda', dtype=torch.bfloat16)
    def forward(self,
                images: List[str],
                texts: List[str],
                question_template: str=default_question_template,
                answer_template: str=default_answer_template) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        """
        assert len(images) == len(texts), "Number of images and texts must match"
        # Turn "a photo of a dog" into
        # Q: "Question: Does this figure show "a photo of a dog"? Please answer yes or no."
        # A: "yes"
        questions = [question_template.format(text) for text in texts]
        answers = [answer_template.format(text) for text in texts]
        
        images = self.load_images(images)

        image_feat = self.model.ln_vision(self.model.visual_encoder(images))
        image_att = torch.ones(image_feat.size()[:-1], dtype=torch.long).to(self.device)
        query_token = self.model.query_tokens.expand(image_feat.shape[0], -1, -1)
        query_att = torch.ones(query_token.size()[:-1], dtype=torch.long).to(query_token.device)

        question_text_Qformer = self.model.tokenizer(
            questions,
            padding='longest',
            truncation=True,
            max_length=self.model.max_txt_len,
            return_tensors="pt",
        ).to(query_token.device)
        Qformer_atts = torch.cat([query_att, question_text_Qformer.attention_mask],dim=1)

        query_output = self.model.Qformer.bert(
            question_text_Qformer.input_ids,
            attention_mask=Qformer_atts,
            query_embeds=query_token,
            encoder_hidden_states=image_feat,
            encoder_attention_mask=image_att,
            return_dict=True,
        )
        
        t5_input = self.model.t5_proj(query_output.last_hidden_state[:,:query_token.size(1),:])
        t5_att = torch.ones(t5_input.size()[:-1], dtype=torch.long).to(query_token.device)

        question_text_t5 = self.model.t5_tokenizer(
            questions,
            padding="longest",
            truncation=True,
            max_length=self.model.max_txt_len,
            return_tensors="pt",
        ).to(query_token.device)
        answer_text_t5 = self.model.t5_output_tokenizer(
            answers,
            padding="longest",
            truncation=True,
            max_length=self.model.max_output_txt_len,
            return_tensors="pt",
        ).to(query_token.device)
        
        encoder_atts = torch.cat([t5_att, question_text_t5.attention_mask], dim=1)

        labels = answer_text_t5.input_ids.masked_fill(
            answer_text_t5.input_ids == self.model.t5_tokenizer.pad_token_id, -100
        )

        inputs_embeds = self.model.t5_model.encoder.embed_tokens(question_text_t5.input_ids)
        inputs_embeds = torch.cat([t5_input, inputs_embeds], dim=1)
        outputs = self.model.t5_model(
            inputs_embeds=inputs_embeds,
            attention_mask=encoder_atts,
            decoder_attention_mask=answer_text_t5.attention_mask,
            return_dict=True,
            labels=labels,
        )
        
        loss_fct = torch.nn.CrossEntropyLoss(reduction='mean')
        lm_prob = torch.zeros(outputs.logits.shape[0])
        for k in range(lm_prob.shape[0]):
            lm_prob[k] = (-loss_fct(outputs.logits[k], labels[k])).exp()
        return lm_prob
    
    @torch.no_grad()
    @torch.autocast(device_type='cuda', dtype=torch.bfloat16)
    def generate(self,
                images: List[str],
                texts: List[str],
                max_length: int = 256) -> List[str]:
        """Forward pass of the model to generate text responses for n (image, text) pairs
        """
        assert len(images) == len(texts), "Number of images and texts must match"
        questions = texts
        
        images = self.load_images(images)

        image_feat = self.model.ln_vision(self.model.visual_encoder(images))
        image_att = torch.ones(image_feat.size()[:-1], dtype=torch.long).to(self.device)
        query_token = self.model.query_tokens.expand(image_feat.shape[0], -1, -1)
        query_att = torch.ones(query_token.size()[:-1], dtype=torch.long).to(query_token.device)

        question_text_Qformer = self.model.tokenizer(
            questions,
            padding='longest',
            truncation=True,
            max_length=self.model.max_txt_len,
            return_tensors="pt",
        ).to(query_token.device)
        Qformer_atts = torch.cat([query_att, question_text_Qformer.attention_mask], dim=1)

        query_output = self.model.Qformer.bert(
            question_text_Qformer.input_ids,
            attention_mask=Qformer_atts,
            query_embeds=query_token,
            encoder_hidden_states=image_feat,
            encoder_attention_mask=image_att,
            return_dict=True,
        )
        
        t5_input = self.model.t5_proj(query_output.last_hidden_state[:,:query_token.size(1),:])
        t5_att = torch.ones(t5_input.size()[:-1], dtype=torch.long).to(query_token.device)

        question_text_t5 = self.model.t5_tokenizer(
            questions,
            padding="longest",
            truncation=True,
            max_length=self.model.max_txt_len,
            return_tensors="pt",
        ).to(query_token.device)
        
        encoder_atts = torch.cat([t5_att, question_text_t5.attention_mask], dim=1)
        inputs_embeds = self.model.t5_model.encoder.embed_tokens(question_text_t5.input_ids)
        inputs_embeds = torch.cat([t5_input, inputs_embeds], dim=1)

        outputs = self.model.t5_model.generate(
            inputs_embeds=inputs_embeds,
            attention_mask=encoder_atts,
            max_length=max_length,
            num_beams=5,
            min_length=1,
            length_penalty=1,
            num_return_sequences=1,
            do_sample=False
        )

        # Decode the generated outputs
        generated_texts = []
        for output in outputs:
            text = self.model.t5_tokenizer.decode(output, skip_special_tokens=True).strip()
            generated_texts.append(text)

        return generated_texts